#!/usr/bin/env python
# -*- coding: utf-8 -*-

# db.dicom_parser.py
"""Parse DICOM RT Dose, Structure, and Plan files for the DVH Analytics SQL Database"""
# Copyright (c) 2016-2019 Dan Cutright
# This file is part of DVH Analytics, released under a BSD license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVH-Analytics

from dicompylercore import dvhcalc, dvh as dicompyler_dvh
from dicompylercore.dicomparser import DicomParser as dicompylerParser
from datetime import datetime
from dateutil.relativedelta import relativedelta  # python-dateutil
from dateutil.parser import parse as date_parser
import numpy as np
from os.path import basename, join
from pubsub import pub
import pydicom
from dvha.options import Options
from dvha.tools.roi_name_manager import clean_name, DatabaseROIs
from dvha.tools.utilities import change_angle_origin, calc_stats, is_date, validate_transfer_syntax_uid
from dvha.tools.roi_formatter import dicompyler_roi_coord_to_db_string, get_planes_from_string
from dvha.tools import roi_geometry as roi_calc
from dvha.tools.mlc_analyzer import Beam as mlca
from dvha.db.sql_connector import DVH_SQL


class DICOM_Parser:
    """
    Parse a set of DICOM files for database
    """
    def __init__(self, plan_file=None, structure_file=None, dose_file=None, dose_sum_file=None, plan_over_rides=None,
                 global_plan_over_rides=None, roi_over_ride=None, roi_map=None, use_dicom_dvh=False, plan_ptvs=None):
        """
        :param plan_file: absolute path of DICOM RT Plan file
        :type plan_file: str
        :param structure_file: absolute path of DICOM RT Struct file
        :type structure_file: str
        :param dose_file: absolute path of DICOM RT Dose file
        :type dose_file: str
        :param dose_sum_file: DICOM RT Dose file with summed dose
        :type dose_sum_file: str
        :param plan_over_rides: Values from import GUI to override DICOM file data
        :type plan_over_rides: dict
        :param global_plan_over_rides: Values from import GUI to override DICOM file data
        :type global_plan_over_rides: dict
        :param roi_over_ride: lookup up for roi name and type overrides
        :type roi_over_ride: dict
        :param roi_map: roi name map
        :type roi_map: DatabaseROIs
        :param use_dicom_dvh: use the DVH stored in DICOM RT-Dose if it exists
        :type use_dicom_dvh: bool
        :param plan_ptvs: assign specifc PTVs for distance calculations
        :type plan_ptvs: list
        """

        self.database_rois = DatabaseROIs() if roi_map is None else roi_map
        self.import_path = Options().IMPORTED_DIR

        self.plan_file = plan_file
        self.structure_file = structure_file
        self.dose_file = dose_file
        self.dose_sum_file = dose_sum_file

        self.use_dicom_dvh = use_dicom_dvh

        # store these values when clearing file loaded data
        self.stored_values = {}

        self.rt_data = {key: None for key in ['plan', 'structure', 'dose']}
        self.dicompyler_data = {key: None for key in ['plan', 'structure', 'dose']}
        self.dicompyler_rt_plan = None
        self.structure_name_and_type = None
        if self.plan_file:
            self.dicompyler_data['plan'] = dicompylerParser(self.plan_file)
            self.rt_data['plan'] = self.dicompyler_data['plan'].ds
            self.dicompyler_rt_plan = self.dicompyler_data['plan'].GetPlan()
        if self.structure_file:
            self.dicompyler_data['structure'] = dicompylerParser(self.structure_file)
            self.rt_data['structure'] = self.dicompyler_data['structure'].ds
            self.structure_name_and_type = self.get_structure_name_and_type(self.rt_data['structure'])
        if self.dose_file:
            if self.dose_sum_file is None:
                # self.rt_data['dose'] = dicompylerParser(self.dose_file).ds
                # The above line may lead to OSError: Deferred read -- original filename not stored. Cannot re-open
                # switching back to pydicom.read_file()
                self.rt_data['dose'] = pydicom.read_file(self.dose_file, force=True)
            else:
                self.rt_data['dose'] = pydicom.read_file(self.dose_sum_file, force=True)

        # These properties are not inherently stored in Pinnacle DICOM files, but can be extracted from dummy ROI
        # names automatically generated by the Pinnacle Script provided by DVH Analytics
        self.poi_rx_data = self.get_rx_data_from_dummy_rois()
        self.poi_tx_site = self.get_tx_site_from_dummy_rois()

        self.__initialize_rx_beam_and_ref_beam_data()

        # over ride objects
        keys = ['mrn', 'study_instance_uid', 'birth_date', 'sim_study_date', 'physician', 'tx_site', 'rx_dose']
        self.plan_over_rides = [plan_over_rides, {key: None for key in keys}][plan_over_rides is None]
        self.global_plan_over_rides = global_plan_over_rides

        self.roi_over_ride = roi_over_ride if roi_over_ride is not None else {'name': {}, 'type': {}}
        for over_ride_type, over_ride in self.roi_over_ride.items():
            for key, value in over_ride.items():
                self.structure_name_and_type[key][over_ride_type] = value

        self.plan_ptvs = plan_ptvs

    def update_stored_values(self):
        keys = ['study_instance_uid_to_be_imported', 'patient_name', 'mrn', 'sim_study_date', 'birth_date',
                'rx_dose', 'ptv_names', 'physician', 'ptv_exists', 'tx_site', 'patient_orientation']
        self.stored_values = {key: getattr(self, key) for key in keys}

    def __initialize_rx_beam_and_ref_beam_data(self):
        beam_num = 0
        self.rx_data = []
        self.beam_data = {}
        self.ref_beam_data = []
        for fx_grp_index, fx_grp_seq in enumerate(self.rt_data['plan'].FractionGroupSequence):
            self.rx_data.append(RxParser(self.rt_data['plan'], self.dicompyler_rt_plan, self.rt_data['structure'],
                                         fx_grp_index, self.poi_rx_data, self.study_instance_uid_to_be_imported))
            self.beam_data[fx_grp_index] = []
            for fx_grp_beam in range(int(fx_grp_seq.NumberOfBeams)):
                beam_number = self.beam_sequence[beam_num].BeamNumber
                beam_seq = self.beam_sequence[beam_num]
                cp_seq = self.get_cp_sequence(self.beam_sequence[beam_num])
                ref_beam_seq_index = self.get_referenced_beam_sequence_index(fx_grp_seq, beam_number)
                ref_beam_seq = fx_grp_seq.ReferencedBeamSequence[ref_beam_seq_index]
                self.beam_data[fx_grp_index].append(BeamParser(beam_seq, ref_beam_seq, cp_seq))

                beam_num += 1

    @staticmethod
    def get_referenced_beam_sequence_index(fx_grp_seq, beam_number):
        """
        Determine the index of the referenced beam sequence for a given beam_number. This properly connects BeamSequence
        beams and ReferencedBeamSequence beams.
        :param fx_grp_seq: Index of the fraction group (likely 0 unless TPS allows multiple prescriptions
        :param beam_number: The beam number of the beam of interest within BeamSequence
        :return: the ReferencedBeamSequence index that matches beam_number
        :rtype: int
        """
        for i, ref_beam_seq in enumerate(fx_grp_seq.ReferencedBeamSequence):
            if ref_beam_seq.ReferencedBeamNumber == beam_number:
                return i
        print('ERROR: Failed to find a matching reference beam in '
              'ReferencedBeamSequence for beam number %s' % beam_number)

    def get_rx_data_from_dummy_rois(self):
        """
        DVH Analytics allows for the user to indicate prescription information embedded in POI names. Some TPSs do not
        include prescription information in the exported DICOM Plan file. For example, Pinnacle scripts are provided
        by DVHA which will create these POIs.  The scripts are found in resources/Pinnacle Scripts

        A POI in the format:
            "rx#: <rx name>: <rx dose in cGy> cGy x <fxs> to <norm. %>%: <norm. method>: <norm. object>"
            where # is rx number starting from 1, will allow DVH-Analytics to retrieve rx information.
        :return: data for the Rxs table that can be extracted from POIs
        :rtype: dict
        """

        if self.structure_file:
            struct_seq = self.rt_data['structure'].StructureSetROISequence
            rx_indices = [i for i, roi in enumerate(struct_seq) if roi.ROIName.lower().startswith('rx ')]

            rx_data = {}
            for i in rx_indices:
                roi_name = struct_seq[i].ROIName.lower()
                name_split = roi_name.split(':')
                if len(name_split) > 3:
                    fx_grp_number = int(name_split[0].strip('rx '))
                    fx_grp_name = name_split[1].strip()
                    fx_dose = float(name_split[2].split('cgy')[0]) / 100.
                    fxs = int(name_split[2].split('x ')[1].split(' to')[0])
                    rx_dose = fx_dose * float(fxs)
                    rx_percent = float(name_split[2].strip().split(' ')[5].strip('%'))
                    normalization_method = name_split[3].strip()

                    if len(name_split) > 4:
                        normalization_object = ['plan_max', name_split[4].strip()][normalization_method != 'plan_max']
                    else:
                        normalization_object = None

                    rx_data[fx_grp_number] = {'fx_grp_number': fx_grp_number,
                                              'fx_grp_name': fx_grp_name,
                                              'fx_dose': fx_dose,
                                              'fxs': fxs,
                                              'rx_dose': rx_dose,
                                              'rx_percent': rx_percent,
                                              'normalization_method': normalization_method,
                                              'normalization_object': normalization_object}

            return rx_data

    def get_tx_site_from_dummy_rois(self):
        """
        DVH Analytics allows for the user to indicate treatment information embedded in POI names. By default, DVHA
        assumes the plan name is the tx site. However, if a POI is provided as follows, the tx site provided in the POI
        will be used instead. Pinnacle scripts are provided by DVHA which will create these POIs. The scripts are found
        in resources/Pinnacle Scripts
        A POI in the format:
            "tx: <site>"
            will allow DVH-Analytics to add this site name to the database upon import.
        :return: treatment site name
        :rtype: str
        """

        if self.structure_file:
            struct_seq = self.rt_data['structure'].StructureSetROISequence
            tx_indices = [i for i, roi in enumerate(struct_seq) if roi.ROIName.lower().startswith('tx: ')]

            if tx_indices:
                return struct_seq[tx_indices[0]].ROIName.split('tx: ')[1]

    def get_plan_row(self):
        """
        Get all data needed for a row in the Plans table of the database
        :return: plan row data with the column name as the key and the values are lists in the format
        [value, column variable type]. This object will be passed to DVH_SQL.insert_row
        :rtype: dict
        """

        return {'mrn': [self.mrn, 'text'],
                'study_instance_uid': [self.study_instance_uid_to_be_imported, 'text'],
                'birth_date': [self.birth_date, 'date'],
                'age': [self.age, 'smallint'],
                'patient_sex': [self.patient_sex, 'char(1)'],
                'sim_study_date': [self.sim_study_date, 'date'],
                'physician': [self.physician, 'varchar(50)'],
                'tx_site': [self.tx_site, 'varchar(50)'],
                'rx_dose': [self.rx_dose, 'real'],
                'fxs': [self.fxs_total, 'int'],
                'patient_orientation': [self.patient_orientation, 'varchar(3)'],
                'plan_time_stamp': [self.plan_time_stamp, 'timestamp'],
                'struct_time_stamp': [self.struct_time_stamp, 'timestamp'],
                'dose_time_stamp': [self.dose_time_stamp, 'timestamp'],
                'tps_manufacturer': [self.tps_manufacturer, 'varchar(50)'],
                'tps_software_name': [self.tps_software_name, 'varchar(50)'],
                'tps_software_version': [self.tps_software_version, 'varchar(30)'],
                'tx_modality': [self.tx_modality, 'varchar(30)'],
                'tx_time': [self.tx_time, 'time'],
                'total_mu': [self.total_mu, 'real'],
                'dose_grid_res': [self.dose_grid_res, 'varchar(16)'],
                'heterogeneity_correction': [self.heterogeneity_correction, ' varchar(30)'],
                'baseline': [None, 'boolean'],
                'import_time_stamp': [None, 'timestamp'],
                'protocol': [None, 'text'],
                'toxicity_grades': [None, 'text'],
                'complexity': [self.plan_complexity, 'real']}

    def get_beam_rows(self):
        """
        Get all beam rows across all prescriptions
        :return: a complete list of beam rows
        :rtype: list
        """
        beam_rows = []
        for rx_index, beam_set in self.beam_data.items():
            for beam in beam_set:
                if float(beam.beam_mu) > 0:  # Ignores set-up fields
                    beam_rows.append(self.get_beam_row(rx_index, beam))
        return beam_rows

    def get_beam_row(self, rx_index, beam):
        """
        Get all data needed for a row in the Beams table of the database
        :param rx_index: the index of self.rx_data associated with beam
        :type rx_index: int
        :param beam: the beam data object to be parsed
        :type beam: BeamParser
        :return: beam row data with the column name as the key and the values are lists in the format
        [value, column variable type]. This object will be passed to DVH_SQL.insert_row
        :rtype: dict
        """
        rx = self.rx_data[rx_index]

        # store these getters so code is repeated every reference
        gantry_values = beam.gantry_values
        collimator_values = beam.collimator_values
        couch_values = beam.couch_values
        mlc_stat_data = beam.mlc_stat_data

        return {'mrn': [self.mrn, 'text'],
                'study_instance_uid': [self.study_instance_uid_to_be_imported, 'text'],
                'beam_number': [beam.beam_number, 'int'],
                'beam_name': [beam.beam_name, 'varchar(30)'],
                'fx_grp_number': [rx.fx_grp_number, 'smallint'],
                'fx_count': [rx.fx_count, 'int'],
                'fx_grp_beam_count': [rx.beam_count, 'smallint'],
                'beam_dose': [beam.beam_dose, 'real'],
                'beam_mu': [beam.beam_mu, 'real'],
                'radiation_type': [beam.radiation_type, 'varchar(30)'],
                'beam_energy_min': [beam.energy_min, 'real'],
                'beam_energy_max': [beam.energy_max, 'real'],
                'beam_type': [beam.beam_type, 'varchar(30)'],
                'control_point_count': [beam.control_point_count, 'int'],
                'gantry_start': [gantry_values['start'], 'real'],
                'gantry_end': [gantry_values['end'], 'real'],
                'gantry_rot_dir': [gantry_values['rot_dir'], 'varchar(5)'],
                'gantry_range': [gantry_values['range'], 'real'],
                'gantry_min': [gantry_values['min'], 'real'],
                'gantry_max': [gantry_values['max'], 'real'],
                'collimator_start': [collimator_values['start'], 'real'],
                'collimator_end': [collimator_values['end'], 'real'],
                'collimator_rot_dir': [collimator_values['rot_dir'], 'varchar(5)'],
                'collimator_range': [collimator_values['range'], 'real'],
                'collimator_min': [collimator_values['min'], 'real'],
                'collimator_max': [collimator_values['max'], 'real'],
                'couch_start': [couch_values['start'], 'real'],
                'couch_end': [couch_values['end'], 'real'],
                'couch_rot_dir': [couch_values['rot_dir'], 'varchar(5)'],
                'couch_range': [couch_values['range'], 'real'],
                'couch_min': [couch_values['min'], 'real'],
                'couch_max': [couch_values['max'], 'real'],
                'beam_dose_pt': [beam.beam_dose_pt, 'varchar(35)'],
                'isocenter': [beam.isocenter, 'varchar(35)'],
                'ssd': [beam.ssd, 'real'],
                'treatment_machine': [beam.treatment_machine, 'varchar(30)'],
                'scan_mode': [beam.scan_mode, 'varchar(30)'],
                'scan_spot_count': [beam.scan_spot_count, 'real'],
                'beam_mu_per_deg': [beam.beam_mu_per_deg, 'real'],
                'beam_mu_per_cp': [beam.beam_mu_per_cp, 'real'],
                'import_time_stamp': [None, 'timestamp'],
                'area_min': [mlc_stat_data['area'][5], 'real'],
                'area_mean': [mlc_stat_data['area'][3], 'real'],
                'area_median': [mlc_stat_data['area'][2], 'real'],
                'area_max': [mlc_stat_data['area'][0], 'real'],
                'perim_min': [mlc_stat_data['area'][5], 'real'],
                'perim_mean': [mlc_stat_data['area'][3], 'real'],
                'perim_median': [mlc_stat_data['area'][2], 'real'],
                'perim_max': [mlc_stat_data['area'][0], 'real'],
                'x_perim_min': [mlc_stat_data['x_perim'][5], 'real'],
                'x_perim_mean': [mlc_stat_data['x_perim'][3], 'real'],
                'x_perim_median': [mlc_stat_data['x_perim'][2], 'real'],
                'x_perim_max': [mlc_stat_data['x_perim'][0], 'real'],
                'y_perim_min': [mlc_stat_data['y_perim'][5], 'real'],
                'y_perim_mean': [mlc_stat_data['y_perim'][3], 'real'],
                'y_perim_median': [mlc_stat_data['y_perim'][2], 'real'],
                'y_perim_max': [mlc_stat_data['y_perim'][0], 'real'],
                'complexity_min': [mlc_stat_data['cmp_score'][5], 'real'],
                'complexity_mean': [mlc_stat_data['cmp_score'][3], 'real'],
                'complexity_median': [mlc_stat_data['cmp_score'][2], 'real'],
                'complexity_max': [mlc_stat_data['cmp_score'][0], 'real'],
                'cp_mu_min': [mlc_stat_data['cp_mu'][5], 'real'],
                'cp_mu_mean': [mlc_stat_data['cp_mu'][3], 'real'],
                'cp_mu_median': [mlc_stat_data['cp_mu'][2], 'real'],
                'cp_mu_max': [mlc_stat_data['cp_mu'][0], 'real'],
                'complexity': [mlc_stat_data['complexity'], 'real'],
                'tx_modality': [beam.tx_modality, 'varchar(35)']}

    def get_rx_rows(self):
        """
        Get all rx rows
        :return: a complete list of rx data objects
        :rtype: list
        """
        return [self.get_rx_row(rx) for rx in self.rx_data]

    def get_rx_row(self, rx):
        """
        Get all data needed for a row in the Rxs table of the database
        :param rx: a rx data object
        :type rx: RxParser
        :return: rx row data with the column name as the key and the values are lists in the format
        [value, column variable type]. This object will be passed to DVH_SQL.insert_row
        :rtype: dict
        """

        data = {'mrn': [self.mrn, 'text'],
                'study_instance_uid': [self.study_instance_uid_to_be_imported, 'text'],
                'plan_name': [rx.plan_name, 'varchar(50)'],
                'fx_grp_name': [rx.fx_grp_name, 'varchar(30)'],
                'fx_grp_number': [rx.fx_grp_number, 'smallint'],
                'fx_grp_count': [self.fx_grp_count, 'smallint'],
                'fx_dose': [rx.fx_dose, 'real'],
                'fxs': [rx.fx_count, 'smallint'],
                'rx_dose': [rx.rx_dose, 'real'],
                'rx_percent': [None, 'real'],
                'normalization_method': [rx.normalization_method, 'varchar(30)'],
                'normalization_object': [rx.normalization_object, 'varchar(30)'],
                'import_time_stamp': [None, 'timestamp']}

        # over-ride values if dummy roi's are used to store rx data in rt_structure
        if self.poi_rx_data and rx.fx_grp_number in list(self.poi_rx_data):
            for key, value in self.poi_rx_data[rx.fx_grp_number].items():
                data[key][0] = value

        for key, value in self.plan_over_rides.items():
            if key in data and value is not None:
                data[key][0] = value

        return data

    def get_dvh_row(self, dvh_index):
        """
        Get all data needed for a row in the DVHs table of the database
        :param dvh_index: the index of the ROI to be imported
        :type dvh_index: int
        :return: dvh row data with the column name as the key and the values are lists in the format
        [value, column variable type]. This object will be passed to DVH_SQL.insert_row
        :rtype: dict
        """

        dvh = None
        if self.use_dicom_dvh:
            try:
                dvh = dicompyler_dvh.DVH.from_dicom_dvh(self.rt_data['dose'], dvh_index)
            except AttributeError:  # dicompyler-core raises this is structure is not found in DICOM DVH
                pass

        if dvh is None:
            try:
                dvh = dvhcalc.get_dvh(self.rt_data['structure'], self.rt_data['dose'], dvh_index,
                                      callback=self.send_dvh_progress)
            except AttributeError:
                dose = validate_transfer_syntax_uid(self.rt_data['dose'])
                structure = validate_transfer_syntax_uid(self.rt_data['structure'])
                dvh = dvhcalc.get_dvh(structure, dose, dvh_index,
                                      callback=self.send_dvh_progress)

        if dvh and dvh.volume > 0:  # ignore points and empty ROIs
            geometries = self.get_dvh_geometries(dvh_index)

            return {'mrn': [self.mrn, 'text'],
                    'study_instance_uid': [self.study_instance_uid_to_be_imported, 'text'],
                    'institutional_roi': [self.get_institutional_roi(dvh_index), 'varchar(50)'],
                    'physician_roi': [self.get_physician_roi(dvh_index), 'varchar(50)'],
                    'roi_name': [self.get_roi_name(dvh_index), 'varchar(50)'],
                    'roi_type': [self.get_roi_type(dvh_index), 'varchar(20)'],
                    'volume': [dvh.volume, 'real'],
                    'min_dose': [dvh.min, 'real'],
                    'mean_dose': [dvh.mean, 'real'],
                    'max_dose': [dvh.max, 'real'],
                    'dvh_string': [','.join(['%.2f' % num for num in dvh.counts]), 'text'],
                    'roi_coord_string': [geometries['roi_coord_str'], 'text'],
                    'dist_to_ptv_min': [None, 'real'],
                    'dist_to_ptv_mean': [None, 'real'],
                    'dist_to_ptv_median': [None, 'real'],
                    'dist_to_ptv_max': [None, 'real'],
                    'surface_area': [geometries['surface_area'], 'real'],
                    'ptv_overlap': [None, 'real'],
                    'import_time_stamp': [None, 'timestamp'],
                    'centroid': [geometries['centroid'], 'varchar(35)'],
                    'dist_to_ptv_centroids': [None, 'real'],
                    'dth_string': [None, 'text'],
                    'spread_x': [geometries['spread'][0], 'real'],
                    'spread_y': [geometries['spread'][1], 'real'],
                    'spread_z': [geometries['spread'][2], 'real'],
                    'cross_section_max': [geometries['cross_sections']['max'], 'real'],
                    'cross_section_median': [geometries['cross_sections']['median'], 'real'],
                    'toxicity_grade': [None, 'smallint']}

    def get_dicom_file_row(self):
        """
        Get all data needed for a row in the DICOM_Files table of the database
        :return: dicom file data with the column name as the key and the values are lists in the format
        [value, column variable type]. This object will be passed to DVH_SQL.insert_row
        :rtype: dict
        """
        return {'mrn': [self.mrn, 'text'],
                'study_instance_uid': [self.study_instance_uid_to_be_imported, 'text'],
                'folder_path': [str(join(self.import_path, self.mrn)), 'text'],
                'plan_file': [basename(self.plan_file), 'text'],
                'structure_file': [basename(self.structure_file), 'text'],
                'dose_file': [basename(self.dose_file), 'text'],
                'import_time_stamp': [None, 'timestamp']}

    @property
    def mrn(self):
        if self.plan_over_rides['mrn'] is not None:
            return self.plan_over_rides['mrn']
        return self.rt_data['plan'].PatientID

    @property
    def study_instance_uid(self):
        return self.rt_data['plan'].StudyInstanceUID

    @property
    def study_instance_uid_to_be_imported(self):
        if hasattr(self, 'plan_over_rides') and self.plan_over_rides['study_instance_uid']:
            return self.plan_over_rides['study_instance_uid']
        return self.rt_data['plan'].StudyInstanceUID

    @property
    def is_study_instance_uid_valid(self):
        return self.is_uid_valid(self.study_instance_uid)

    @property
    def is_study_instance_uid_to_be_imported_valid(self):
        return self.is_uid_valid(self.study_instance_uid_to_be_imported)

    @staticmethod
    def is_uid_valid(uid):
        valid = False
        if uid:
            with DVH_SQL() as cnx:
                valid = not cnx.is_study_instance_uid_in_table('Plans', uid)
        return valid

    @property
    def patient_name(self):
        return str(self.rt_data['plan'].PatientName)

    @property
    def file_set_complete(self):
        return all([self.plan_file, self.structure_file, self.dose_file])

    @property
    def missing_files(self):
        return [key.capitalize() for key, value in self.rt_data.items() if value is None]

    # ------------------------------------------------------------------------------
    # Plan table data
    # ------------------------------------------------------------------------------
    @property
    def tx_modality(self):
        tx_modalities = []
        for fx_grp_index, beam_parser_list in self.beam_data.items():
            for beam_parser in beam_parser_list:
                tx_modalities.append(beam_parser.tx_modality)
        tx_modalities = sorted(list(set(tx_modalities)))
        return ','.join(tx_modalities)

    @property
    def rx_dose(self):
        if self.plan_over_rides['rx_dose'] is not None:
            ans = self.plan_over_rides['rx_dose']
        elif self.poi_rx_data:
            ans = sum([rx['rx_dose'] for rx in self.poi_rx_data.values()])
        else:
            ans = sum([rx.rx_dose for rx in self.rx_data if rx.rx_dose])
        return self.process_global_over_ride('rx_dose', ans)

    @property
    def total_mu(self):
        mus = []
        for fx_grp_index, beam_parser_list in self.beam_data.items():
            for beam_parser in beam_parser_list:
                if self.rx_data[fx_grp_index].fx_count:
                    if beam_parser.beam_mu:
                        mus.append(beam_parser.beam_mu * self.rx_data[fx_grp_index].fx_count)
        return sum(mus)

    @property
    def heterogeneity_correction(self):
        if hasattr(self.rt_data['dose'], 'TissueHeterogeneityCorrection'):
            if isinstance(self.rt_data['dose'].TissueHeterogeneityCorrection, str):
                heterogeneity_correction = self.rt_data['dose'].TissueHeterogeneityCorrection
            else:
                heterogeneity_correction = ','.join(self.rt_data['dose'].TissueHeterogeneityCorrection)
        else:
            heterogeneity_correction = 'IMAGE'

        return heterogeneity_correction

    @property
    def patient_sex(self):
        return self.get_attribute('plan', 'PatientSex')

    @property
    def sim_study_date(self):
        if self.plan_over_rides['sim_study_date'] is not None:
            ans = self.plan_over_rides['sim_study_date']
        else:
            ans = self.get_attribute('plan', 'StudyDate')
        return self.process_global_over_ride('sim_study_date', ans)

    @property
    def birth_date(self):
        if self.plan_over_rides['birth_date'] is not None:
            ans = self.plan_over_rides['birth_date']
        else:
            ans = self.get_attribute('plan', 'PatientBirthDate')
        return self.process_global_over_ride('birth_date', ans)

    @property
    def age(self):
        if self.sim_study_date and self.birth_date:
            try:
                dates = {'sim_study_date': None, 'birth_date': None}
                for date_type in list(dates):
                    dates[date_type] = getattr(self, date_type)
                    if type(dates[date_type]) is int or float:
                        dates[date_type] = str(dates[date_type])
                    if type(dates[date_type]) is str:
                        dates[date_type] = date_parser(dates[date_type]).date()
                age = relativedelta(dates['sim_study_date'], dates['birth_date']).years
                if age >= 0:
                    return age
            except Exception:
                pass

    @property
    def physician(self):
        if self.plan_over_rides['physician'] is not None:
            ans = self.plan_over_rides['physician']
        else:
            ans = str(self.get_attribute('plan', ['PhysiciansOfRecord', 'ReferringPhysicianName'])).upper()
            if not ans:
                ans = 'DEFAULT'
        return self.process_global_over_ride('physician', ans)

    @property
    def fxs(self):
        try:
            fx_grp_seq = self.rt_data['plan'].FractionGroupSequence
            return [int(float(fx_grp.NumberOfFractionsPlanned)) for fx_grp in fx_grp_seq]
        except ValueError:
            pass

    @property
    def fxs_total(self):
        fxs = self.fxs
        if fxs:
            return sum(fxs)

    @property
    def fx_grp_count(self):
        return len(self.rt_data['plan'].FractionGroupSequence)

    @property
    def patient_orientation(self):
        # TODO: database assumes only one orientation (i.e., three characters)
        seq = self.get_attribute('plan', 'PatientSetupSequence')
        if seq is not None:
            # return ','.join([setup.PatientPosition for setup in seq])
            return str(seq[0].PatientPosition)[:3]

    @property
    def plan_time_stamp(self):
        return self.get_time_stamp('plan')

    @property
    def struct_time_stamp(self):
        return self.get_time_stamp('structure')

    @property
    def dose_time_stamp(self):
        return self.get_time_stamp('dose')

    @property
    def tps_manufacturer(self):
        return self.get_attribute('plan', 'Manufacturer')

    @property
    def tps_software_name(self):
        return self.get_attribute('plan', 'ManufacturerModelName')

    @property
    def tps_software_version(self):
        # Some TPSs may store the version in RT Dose rather than plan
        # SoftwareVersions may also be stored as a string rather than a list
        for rt_type in ['plan', 'dose', 'structure']:  # Check each rt_type until SoftwareVersions is found
            version = self.get_attribute(rt_type, 'SoftwareVersions')
            if version is not None:
                if isinstance(version, str):
                    version = [version]
                return ','.join(version)

    @property
    def tx_site(self):
        if self.plan_over_rides['tx_site'] is not None:
            ans = self.plan_over_rides['tx_site']
        elif self.poi_tx_site is not None:
            ans = self.poi_tx_site
        else:
            ans = self.get_attribute('plan', 'RTPlanLabel')
        return self.process_global_over_ride('tx_site', ans)

    @property
    def brachy(self):
        return hasattr(self.rt_data['plan'], 'BrachyTreatmentType')

    @property
    def brachy_type(self):
        return self.get_attribute('plan', 'BrachyTreatmentType')

    @property
    def proton(self):
        return hasattr(self.rt_data['plan'], 'IonBeamSequence')

    @property
    def beam_sequence(self):
        if hasattr(self.rt_data['plan'], 'BeamSequence'):
            return self.rt_data['plan'].BeamSequence
        elif hasattr(self.rt_data['plan'], 'IonBeamSequence'):
            return self.rt_data['plan'].IonBeamSequence

    def get_cp_sequence(self, beam_seq):
        if hasattr(beam_seq, 'ControlPointSequence'):
            return beam_seq.ControlPointSequence
        elif hasattr(beam_seq, 'IonControlPointSequence'):
            return beam_seq.IonControlPointSequence

    @property
    def photon(self):
        return self.is_beam_with_rad_type_in_plan('photon')

    @property
    def electron(self):
        return self.is_beam_with_rad_type_in_plan('electron')

    @property
    def radiation_type(self):
        rad_types = {'PHOTONS': self.photon,
                     'ELECTRONS': self.electron,
                     'PROTONS': self.proton,
                     'BRACHY': self.brachy_type}
        types = sorted([rad_type for rad_type, rad_value in rad_types.items() if rad_value])
        return ','.join(types)

    @property
    def tx_time(self):
        if hasattr(self.rt_data['plan'], 'BrachyTreatmentType') and \
                hasattr(self.rt_data['plan'], 'ApplicationSetupSequence'):
            seconds = 0
            for app_seq in self.rt_data['plan'].ApplicationSetupSequence:
                for chan_seq in app_seq.ChannelSequence:
                    seconds += chan_seq.ChannelTotalTime
            m, s = divmod(seconds, 60)
            h, m = divmod(m, 60)
            return "%02d:%02d:%02d" % (h, m, s)

    @property
    def dose_grid_res(self):
        try:
            dose_grid_resolution = [str(round(float(self.rt_data['dose'].PixelSpacing[0]), 1)),
                                    str(round(float(self.rt_data['dose'].PixelSpacing[1]), 1))]
            if hasattr(self.rt_data['dose'], 'SliceThickness') and self.rt_data['dose'].SliceThickness:
                dose_grid_resolution.append(str(round(float(self.rt_data['dose'].SliceThickness), 1)))
            return ', '.join(dose_grid_resolution)
        except Exception:
            pass

    def process_global_over_ride(self, key, pre_over_ride_value):
        if self.global_plan_over_rides:
            over_ride = self.global_plan_over_rides[key]
            if over_ride['value']:
                if not over_ride['only_if_missing'] or (over_ride['only_if_missing'] and not pre_over_ride_value):
                    return over_ride['value']
        return pre_over_ride_value

    @property
    def plan_complexity(self):
        plan_complexity = 0
        for fx in self.beam_data.values():
            for beam in fx:
                complexity = beam.mlc_stat_data['complexity']
                if complexity:
                    plan_complexity += complexity
        if plan_complexity:
            return plan_complexity

    # ------------------------------------------------------------------------------
    # DVH table data
    # ------------------------------------------------------------------------------
    def get_roi_type(self, key):
        """
        Get the ROI type as defined in DICOM (e.g., PTV, ORGAN, EXTERNAL, etc.)
        :param key: index of ROI
        :type key: int
        :return: the roi type
        :rtype: str
        """
        if key in list(self.roi_over_ride['type']):
            return self.roi_over_ride['type'][key]
        return self.structure_name_and_type[key]['type'].upper()

    def reset_roi_type_over_ride(self, key):
        self.roi_over_ride['type'][key] = None

    def get_roi_name(self, key):
        """
        :param key: the index of the roi
        :return: roi name to be used in the database
        :rtype: str
        """
        if key in list(self.roi_over_ride['name']):
            return clean_name(self.roi_over_ride['name'][key])
        return clean_name(self.structure_name_and_type[key]['name'])

    def get_physician_roi(self, key):
        """
        Look up the physician roi of the specified roi for this plan
        :param key: the index of the roi
        :return: physician roi or 'uncategorized'
        :rtype: str
        """
        roi_name = self.get_roi_name(key)
        if self.database_rois.is_roi(roi_name):
            if self.database_rois.is_physician(self.physician):
                return self.database_rois.get_physician_roi(self.physician, roi_name)
        return 'uncategorized'

    def get_institutional_roi(self, key):
        """
        Look up the institutional roi of the specified roi for this plan
        :param key: the index of the roi
        :return: institutional roi or 'uncategorized'
        :rtype: str
        """
        roi_name = self.get_roi_name(key)
        if self.database_rois.is_roi(roi_name):
            if self.database_rois.is_physician(self.physician):
                return self.database_rois.get_institutional_roi(self.physician, self.get_physician_roi(key))
            if roi_name in self.database_rois.institutional_rois:
                return roi_name
        return 'uncategorized'

    @property
    def ptv_exists(self):
        """
        Check if the plan has at least one PTV assigned
        :rtype: bool
        """
        if self.structure_name_and_type:
            for key in list(self.structure_name_and_type):
                if self.get_roi_type(key) == 'PTV':
                    return True
        return False

    @property
    def roi_names(self):
        """
        :return: roi_names for this plan per the plan (clean_name not applied)
        :rtype: list
        """
        return [self.get_roi_name(key) for key in list(self.structure_name_and_type)]

    def get_roi_key(self, roi_name):
        """
        Look-up the roi index for a given roi_name
        :param roi_name: the ROI name to be looked-up
        :return: the roi index or None
        :rtype: int
        """
        roi_name = clean_name(roi_name)
        for key in list(self.structure_name_and_type):
            if roi_name == clean_name(self.get_roi_name(key)):
                return key

    @property
    def ptv_names(self):
        """
        Scan all ROIs for PTVs
        :return: PTV names
        :rtype: list
        """
        if self.structure_name_and_type:
            return [self.get_roi_name(x) for x in list(self.structure_name_and_type) if self.get_roi_type(x) == 'PTV']
        return self.stored_values['ptv_names']

    def get_surface_area(self, key):
        """
        :param key: the index of the roi
        :return: the surface area of the ROI in cm^2
        :rtype: float
        """
        coord = self.dicompyler_data['structure'].GetStructureCoordinates(key)
        try:
            return roi_calc.surface_area(coord)
        except Exception:
            print("Surface area calculation failed for key, name: %s, %s" % (key, self.get_roi_name(key)))

    def get_dvh_geometries(self, key):
        """
        Collect ROI geometry information to be stored in the database
        :param key: the index of the roi
        :return: geometric information
        :rtype: dict
        """

        structure_coord = self.dicompyler_data['structure'].GetStructureCoordinates(key)

        try:
            surface_area = roi_calc.surface_area(structure_coord)
        except Exception:
            print("Surface area calculation failed for key, name: %s, %s" % (key, self.get_roi_name(key)))
            surface_area = None

        roi_coord_str = dicompyler_roi_coord_to_db_string(structure_coord)
        planes = get_planes_from_string(roi_coord_str)
        centroid = roi_calc.centroid(planes)
        spread = roi_calc.spread(planes)
        cross_sections = roi_calc.cross_section(planes)

        return {'roi_coord_str': roi_coord_str,
                'surface_area': surface_area,
                'centroid': ','.join([str(round(x, 3)) for x in centroid]),
                'spread': spread,
                'cross_sections': cross_sections}

    # ------------------------------------------------------------------------------
    # Generic tools
    # ------------------------------------------------------------------------------
    @property
    def init_param(self):
        params = ['plan_file', 'structure_file', 'dose_file', 'dose_sum_file',
                  'plan_over_rides', 'global_plan_over_rides', 'roi_over_ride']
        return {key: getattr(self, key) for key in params}

    def get_attribute(self, rt_type, pydicom_attribute):
        """
        :param rt_type: plan. dose, or structure
        :type rt_type: str
        :param pydicom_attribute: attribute as specified in pydicom
        :type pydicom_attribute: str or list of str
        :return: pydicom value or None
        """
        if isinstance(pydicom_attribute, str):
            pydicom_attribute = [pydicom_attribute]

        for attribute in pydicom_attribute:
            if hasattr(self.rt_data[rt_type], attribute):
                return getattr(self.rt_data[rt_type], attribute)

    def get_time_stamp(self, rt_type):
        """
        Get the time stamp as specified in the DICOM file
        :param rt_type: 'plan', 'structure' or 'dose'
        :return: datetime
        :type: datetime
        """

        attribute = {'date': {'plan': 'RTPlanDate',
                              'structure': 'StructureSetDate',
                              'dose': 'InstanceCreationDate'},
                     'time': {'plan': 'RTPlanTime',
                              'structure': 'StructureSetTime',
                              'dose': 'InstanceCreationTime'}}

        datetime_str = self.get_attribute(rt_type, attribute['date'][rt_type])
        time = self.get_attribute(rt_type, attribute['time'][rt_type])

        try:
            if time:
                time = time.split('.')[0]  # ignore fractional sec
                return datetime.strptime(datetime_str + time, '%Y%m%d%H%M%S')
            else:
                return datetime.strptime(datetime_str, '%Y%m%d')
        except Exception as e:
            print(e)

    def is_beam_with_rad_type_in_plan(self, rad_type):
        """
        If a beam is photon or electron, beam data is stored in BeamSequence as opposed to IonBeamSequence
        :param rad_type: the radiation type of the beam (e.g., 'photon' or 'electron')
        :type rad_type: str
        :return: True if any beam in plan is the specified rad_type
        :rtype: bool
        """
        if hasattr(self.rt_data['plan'], 'BeamSequence'):
            for beam in self.rt_data['plan'].BeamSequence:
                if rad_type in beam.RadiationType:
                    return True
        return False

    @property
    def pre_import_data(self):
        """
        :return: Only the data needed for the pre import screen, pass to PreImportData
        :rtype: dict
        """
        self.update_stored_values()
        return {'file_set': {'plan': self.plan_file, 'dose': self.dose_file, 'structure': self.structure_file},
                'stored_values': self.stored_values,
                'dicompyler_rt_structures': self.structure_name_and_type,
                'roi_map': self.database_rois}

    @staticmethod
    def get_structure_name_and_type(rt_struct_ds):
        """Stripped down version of dicompylercore.dicomparser.GetStructures"""
        structures = {}

        # Locate the name and number of each ROI
        if 'StructureSetROISequence' in rt_struct_ds:
            for item in rt_struct_ds.StructureSetROISequence:
                data = {}
                number = int(item.ROINumber)
                data['id'] = number
                data['name'] = item.ROIName
                structures[number] = data

        # Determine the type of each structure (PTV, organ, external, etc)
        if 'RTROIObservationsSequence' in rt_struct_ds:
            for item in rt_struct_ds.RTROIObservationsSequence:
                number = item.ReferencedROINumber
                if number in structures:
                    structures[number]['type'] = item.RTROIInterpretedType

        return structures

    @staticmethod
    def send_dvh_progress(current_plane, plane_count):
        """
        Callback for dicompyler-core dvh calculation
        :param current_plane: the plane to be calculated
        :type current_plane: int
        :param plane_count: total number of planes to be calculated
        :type plane_count: int
        """
        progress = float(current_plane) / float(plane_count)
        pub.sendMessage('update_dvh_progress', msg=progress)


class BeamParser:
    """
    This class is used to parse beam data needed for importing a plan into the database
    """
    def __init__(self, beam_data, ref_beam_data, cp_seq):
        """
        :param beam_data:
        :param ref_beam_data:
        :param cp_seq:
        """
        self.beam_data = beam_data
        self.ref_beam_data = ref_beam_data
        self.cp_seq = cp_seq

    ################################################################################
    # General beam parser tools
    ################################################################################
    @staticmethod
    def get_data_attribute(data_obj, pydicom_attr, default=None, data_type=None):
        """
        Generic method to get the value of a pydicom dataset attribute
        :param data_obj: a pydicom dataset
        :param pydicom_attr: the attribute of interest
        :type pydicom_attr: str
        :param default: if attribute is not found, this value is returned
        :param data_type: set to 'float' or 'int' to convert return value as specified
        :type data_type: str
        :return: the value of the pydicom attribute
        """
        if hasattr(data_obj, pydicom_attr):
            value = getattr(data_obj, pydicom_attr)
            if data_type == 'float':
                return float(value)
            elif data_type == 'int':
                return int(float(value))
            return value
        return default

    def get_point_attribute(self, data_obj, pydicom_attr):
        """
        Generic method to get the value of a pydicom dataset attribute that represents a point
        :param data_obj: a pydicom dataset
        :param pydicom_attr: the attribute of interest
        :return: a csv of the point values
        :rtype: str
        """
        point = self.get_data_attribute(data_obj, pydicom_attr)
        if point:
            return ','.join([str(round(dim_value, 3)) for dim_value in point])

    def get_cp_attributes(self, pydicom_attr, force_float=False, force_int=False):
        """
        Generic method to get the values of a pydicom attribute for self.cp_seqeunce
        This is used anytime you need to retrieve values that are specified for every control point
        :param pydicom_attr: the attribute of interest
        :type pydicom_attr: str
        :param force_float: force values to be of type float
        :type force_float: bool
        :param force_int: force values to eb of type int
        :type force_int: bool
        :return: the values of the specified attribute across all control points
        :rtype: list
        """
        values = []
        for cp in self.cp_seq:
            if hasattr(cp, pydicom_attr):
                if 'Rotation' in pydicom_attr:
                    if getattr(cp, pydicom_attr).upper() in {'CC', 'CW'}:
                        values.append(getattr(cp, pydicom_attr).upper())
                else:
                    value = getattr(cp, pydicom_attr)
                    if force_float:
                        value = float(value)
                    if force_int:
                        value = int(float(value))
                    values.append(value)

        if pydicom_attr[-5:] == 'Angle':
            values = change_angle_origin(values, 180)  # if angle is greater than 180, convert to negative value
        return values

    def get_angle_values(self, angle_type):
        """
        Collect data to be used for database import related to objects that have angles
        :param angle_type: 'gantry', 'collimator' or 'couch'
        :type angle_type: 'str
        :return: start, end, rot_dir, range, min, and max values
        :rtype: dict
        """
        angles = getattr(self, '%s_angles' % angle_type)
        return {'start': angles[0],
                'end': angles[-1],
                'rot_dir': self.get_rotation_direction(getattr(self, '%s_rot_dirs' % angle_type)),
                'range': round(float(np.sum(np.abs(np.diff(angles)))), 1),
                'min': min(angles),
                'max': max(angles)}

    @staticmethod
    def get_rotation_direction(rotation_list):
        """
        Given a list of rotation directions, return either the only direction or both with the initial direction first
        :param rotation_list: rotation angles of a beam attribute
        :return: concise description of rotation
        :rtype: str
        """
        if not rotation_list:
            return None
        if len(set(rotation_list)) == 1:  # Only one direction found
            return rotation_list[0]
        return ['CC/CW', 'CW/CC'][rotation_list[0] == 'CW']

    ################################################################################
    # Properties
    ################################################################################
    @property
    def beam_number(self):
        return int(self.beam_data.BeamNumber)

    @property
    def beam_name(self):
        if hasattr(self.beam_data, 'BeamDescription'):
            return self.beam_data.BeamDescription
        if hasattr(self.beam_data, 'BeamName'):
            return self.beam_data.BeamName
        return self.beam_number

    @property
    def treatment_machine(self):
        return self.get_data_attribute(self.beam_data, 'TreatmentMachineName')

    @property
    def beam_dose(self):
        return self.get_data_attribute(self.ref_beam_data, 'BeamDose', default=0., data_type='float')

    @property
    def beam_mu(self):
        return self.get_data_attribute(self.ref_beam_data, 'BeamMeterset', default=0., data_type='float')

    @property
    def beam_dose_pt(self):
        return self.get_point_attribute(self.ref_beam_data, 'BeamDoseSpecificationPoint')

    @property
    def isocenter(self):
        return self.get_point_attribute(self.cp_seq[0], 'IsocenterPosition')

    @property
    def beam_type(self):
        return self.get_data_attribute(self.beam_data, 'BeamType')

    @property
    def radiation_type(self):
        return self.get_data_attribute(self.beam_data, 'RadiationType')

    @property
    def scan_mode(self):
        return self.get_data_attribute(self.beam_data, 'ScanMode')

    @property
    def scan_spot_count(self):
        if hasattr(self.cp_seq[0], 'NumberOfScanSpotPositions'):
            return sum([int(cp.NumberOfScanSpotPositions) for cp in self.cp_seq]) / 2

    @property
    def energies(self):
        return self.get_cp_attributes('NominalBeamEnergy')

    @property
    def energy_min(self):
        return round(min(self.energies), 2)

    @property
    def energy_max(self):
        return round(max(self.energies), 2)

    @property
    def gantry_angles(self):
        return self.get_cp_attributes('GantryAngle', force_float=True)

    @property
    def collimator_angles(self):
        return self.get_cp_attributes('BeamLimitingDeviceAngle', force_float=True)

    @property
    def couch_angles(self):
        return self.get_cp_attributes('PatientSupportAngle', force_float=True)

    @property
    def gantry_rot_dirs(self):
        return self.get_cp_attributes('GantryRotationDirection', force_float=True)

    @property
    def collimator_rot_dirs(self):
        return self.get_cp_attributes('BeamLimitingDeviceRotationDirection')

    @property
    def couch_rot_dirs(self):
        return self.get_cp_attributes('PatientSupportRotationDirection')

    @property
    def gantry_values(self):
        return self.get_angle_values('gantry')

    @property
    def collimator_values(self):
        return self.get_angle_values('collimator')

    @property
    def couch_values(self):
        return self.get_angle_values('couch')

    @property
    def is_arc(self):
        return bool(self.get_angle_values('gantry')['rot_dir'])

    @property
    def tx_modality(self):
        rad_type = str(self.radiation_type)
        if not rad_type:
            return None
        if 'brachy' in rad_type.lower():
            return rad_type
        return "%s %s" % (self.radiation_type.title(), ['3D', 'Arc'][self.is_arc])

    @property
    def control_point_count(self):
        return int(self.get_data_attribute(self.beam_data, 'NumberOfControlPoints'))

    @property
    def ssd(self):
        ssds = self.get_cp_attributes('SourceToSurfaceDistance', force_float=True)
        if ssds:
            return round(float(np.average(ssds))/10., 2)

    @property
    def beam_mu_per_deg(self):
        try:
            return round(self.beam_mu / self.gantry_values['range'], 2)
        except Exception:
            pass

    @property
    def beam_mu_per_cp(self):
        try:
            return round(self.beam_mu / self.control_point_count, 2)
        except Exception:
            pass

    @property
    def mlc_stat_data(self):
        mlc_keys = ['area', 'x_perim', 'y_perim', 'perim', 'cmp_score', 'cp_mu']
        try:
            mlc_summary_data = mlca(self.beam_data, self.beam_mu, ignore_zero_mu_cp=True).summary
            mlca_stat_data = {key: calc_stats(mlc_summary_data[key]) for key in mlc_keys}
            mlca_stat_data['complexity'] = np.sum(mlc_summary_data['cmp_score'])
        except Exception as e:
            print('WARNING: Skipping mlc_stat_data calculation because ', str(e))
            mlca_stat_data = {key: [None] * 6 for key in mlc_keys}
            mlca_stat_data['complexity'] = None
        return mlca_stat_data


class RxParser:
    """
    This class is used to parse rx data needed for importing a plan into the database
    """
    def __init__(self, rt_plan, dicompyler_plan, rt_structure, fx_grp_index, pinnacle_rx_data, study_instance_uid):
        """
        :param rt_plan: the pydicom dataset object from reading the RT Plan DICOM file
        :param dicompyler_plan: the dicompyler-core object from readying the RT Plan DICOM file
        :param rt_structure: the pydicom dataset object from reading the RT Structure DICOM file
        :param fx_grp_index: the index of the fraction group associated with the rx to be parsed
        :type fx_grp_index: int
        :param pinnacle_rx_data: prescription data extracted from dummy POIs
        :type pinnacle_rx_data: dict
        :param study_instance_uid: The study instance UID of the associated plan.
        :type study_instance_uid: str
        """
        self.rt_plan = rt_plan
        self.dicompyler_plan = dicompyler_plan
        self.rt_structure = rt_structure
        self.study_instance_uid = study_instance_uid
        self.fx_grp_data = rt_plan.FractionGroupSequence[fx_grp_index]
        self.dose_ref_index = self.get_dose_ref_seq_index()

        self.pinnacle_rx_data = None
        if pinnacle_rx_data and fx_grp_index+1 in list(pinnacle_rx_data):
            self.pinnacle_rx_data = pinnacle_rx_data[fx_grp_index+1]

    ################################################################################
    # General rx parser tools
    ################################################################################
    def get_dose_ref_seq_index(self):
        if self.has_dose_ref:
            for i, dose_ref in enumerate(self.rt_plan.DoseReferenceSequence):
                if dose_ref.DoseReferenceNumber == self.fx_grp_number:
                    return i

    def get_dose_ref_attr(self, pydicom_attr):
        if self.has_dose_ref and self.dose_ref_index is not None:
            dose_ref_data = self.rt_plan.DoseReferenceSequence[self.dose_ref_index]
            if hasattr(dose_ref_data, pydicom_attr):
                return getattr(dose_ref_data, pydicom_attr)

    ################################################################################
    # Properties
    ################################################################################
    @property
    def plan_name(self):
        if hasattr(self.rt_plan, 'RTPlanLabel'):
            return self.rt_plan.RTPlanLabel

    @property
    def fx_grp_number(self):
        """
        If there are multiple RT Plan files for a given study instance uid, this function will ensure fx_grp_number
        is unique
        :rtype: int
        """
        with DVH_SQL() as cnx:
            fraction_group_start = cnx.get_max_value('Rxs', 'fx_grp_number',
                                                     "study_instance_uid = '%s'" % self.study_instance_uid)
        if not fraction_group_start:
            fraction_group_start = 0
        return self.fx_grp_data.FractionGroupNumber + fraction_group_start

    @property
    def fx_grp_name(self):
        if self.pinnacle_rx_data:
            return self.pinnacle_rx_data['fx_grp_name']
        return "FxGrp %s" % self.fx_grp_number

    @property
    def has_dose_ref(self):
        return hasattr(self.rt_plan, 'DoseReferenceSequence')

    @property
    def rx_dose(self):
        if self.pinnacle_rx_data:
            return self.pinnacle_rx_data['rx_dose']
        ans = self.get_dose_ref_attr('TargetPrescriptionDose')
        if ans is None:
            ans = float(self.dicompyler_plan['rxdose']) / 100.
        return ans

    @property
    def rx_percentage(self):
        if self.pinnacle_rx_data:
            return self.pinnacle_rx_data['rx_percentage']

    @property
    def fx_count(self):
        return int(self.fx_grp_data.NumberOfFractionsPlanned)

    @property
    def fx_dose(self):
        if self.pinnacle_rx_data:
            return self.pinnacle_rx_data['fx_dose']
        try:
            return round(self.rx_dose / float(self.fx_count), 2)
        except Exception:
            print('WARNING: Unable to calculate fx_dose')

    @property
    def normalization_method(self):
        if self.pinnacle_rx_data:
            return self.pinnacle_rx_data['normalization_method']
        return self.get_dose_ref_attr('DoseReferenceStructureType')

    @property
    def normalization_object(self):
        if self.pinnacle_rx_data:
            return self.pinnacle_rx_data['normalization_object']
        if self.normalization_method:
            if self.normalization_method.lower() == 'coordinates':
                return 'COORDINATE'
            elif self.normalization_method.lower() == 'site':
                if hasattr(self.rt_plan, 'ManufacturerModelName'):
                    return self.rt_plan.ManufacturerModelName

    @property
    def beam_count(self):
        if hasattr(self.fx_grp_data, 'NumberOfBeams'):
            return int(self.fx_grp_data.NumberOfBeams)

    @property
    def beam_numbers(self):
        return [ref_beam.ReferencedBeamNumber for ref_beam in self.fx_grp_data.ReferencedBeamSequence]


class PreImportData:
    """
    Light-weight object for the DICOM Importer GUI to avoid memory allocation issues
    """
    def __init__(self, file_set, stored_values, dicompyler_rt_structures, roi_map=None):

        self.plan_file = file_set['plan']
        self.dose_file = file_set['dose']
        self.structure_file = file_set['structure']
        self.missing_files = [key.capitalize() for key, value in file_set.items() if value is None]

        self.stored_values = stored_values
        self.dicompyler_rt_structures = dicompyler_rt_structures
        self.database_rois = DatabaseROIs() if roi_map is None else roi_map

        # over ride objects
        keys = ['mrn', 'study_instance_uid', 'birth_date', 'sim_study_date', 'physician', 'tx_site', 'rx_dose']
        self.plan_over_rides = {key: None for key in keys}
        self.global_plan_over_rides = None

        self.roi_over_ride = {'name': {}, 'type': {}}

        # If importing with auto sum turned off, use this list to track plan-specific PTVs
        self.plan_ptvs = []

    @property
    def mrn(self):
        if self.plan_over_rides['mrn'] is not None:
            return self.plan_over_rides['mrn']
        return self.stored_values['mrn']

    @property
    def study_instance_uid(self):
        return self.stored_values['study_instance_uid_to_be_imported']

    @property
    def study_instance_uid_to_be_imported(self):
        if hasattr(self, 'plan_over_rides') and self.plan_over_rides['study_instance_uid']:
            return self.plan_over_rides['study_instance_uid']
        return self.stored_values['study_instance_uid_to_be_imported']

    @property
    def is_study_instance_uid_valid(self):
        return self.is_uid_valid(self.study_instance_uid)

    @property
    def is_study_instance_uid_to_be_imported_valid(self):
        return self.is_uid_valid(self.study_instance_uid_to_be_imported)

    @staticmethod
    def is_uid_valid(uid):
        valid = False
        if uid:
            with DVH_SQL() as cnx:
                valid = not cnx.is_study_instance_uid_in_table('Plans', uid)
        return valid

    @property
    def patient_name(self):
        return self.stored_values['patient_name']

    @property
    def file_set_complete(self):
        return all([self.plan_file, self.structure_file, self.dose_file])

    @property
    def rx_dose(self):
        if self.plan_over_rides['rx_dose'] is not None:
            ans = self.plan_over_rides['rx_dose']
        else:
            ans = self.stored_values['rx_dose']
        return self.process_global_over_ride('rx_dose', ans)

    @property
    def sim_study_date(self):
        if self.plan_over_rides['sim_study_date'] is not None:
            ans = self.plan_over_rides['sim_study_date']
        else:
            ans = self.stored_values['sim_study_date']
        return self.process_global_over_ride('sim_study_date', ans)

    @property
    def birth_date(self):
        if self.plan_over_rides['birth_date'] is not None:
            ans = self.plan_over_rides['birth_date']
        else:
            ans = self.stored_values['birth_date']
        return self.process_global_over_ride('birth_date', ans)

    @property
    def physician(self):
        if self.plan_over_rides['physician'] is not None:
            ans = self.plan_over_rides['physician']
        else:
            ans = self.stored_values['physician']
        return self.process_global_over_ride('physician', ans)

    @property
    def tx_site(self):
        if self.plan_over_rides['tx_site'] is not None:
            ans = self.plan_over_rides['tx_site']
        else:
            ans = self.stored_values['tx_site']
        return self.process_global_over_ride('tx_site', ans)

    @property
    def patient_orientation(self):
        return self.stored_values['patient_orientation']

    def process_global_over_ride(self, key, pre_over_ride_value):
        if self.global_plan_over_rides:
            over_ride = self.global_plan_over_rides[key]
            if over_ride['value']:
                if not over_ride['only_if_missing'] or (over_ride['only_if_missing'] and not pre_over_ride_value):
                    return over_ride['value']
        return pre_over_ride_value

    def get_roi_type(self, key):
        """
        Get the ROI type as defined in DICOM (e.g., PTV, ORGAN, EXTERNAL, etc.)
        :param key: index of ROI
        :type key: int
        :return: the roi type
        :rtype: str
        """
        if key in list(self.roi_over_ride['type']):
            ans = self.roi_over_ride['type'][key]
        else:
            ans = self.dicompyler_rt_structures[key]['type'].upper()
        return ans if ans else 'NONE'

    def reset_roi_type_over_ride(self, key):
        self.roi_over_ride['type'][key] = None

    def get_roi_name(self, key):
        """
        Applies clean_name from roi_name_manager.py to the name in the DICOM file
        :param key: the index of the roi
        :return: roi name to be used in the database
        :rtype: str
        """
        if key in list(self.roi_over_ride['name']):
            return clean_name(self.roi_over_ride['name'][key])
        return clean_name(self.dicompyler_rt_structures[key]['name'])

    def set_roi_name(self, key, name):
        self.roi_over_ride['name'][key] = clean_name(name)

    def get_physician_roi(self, key):
        """
        Look up the physician roi of the specified roi for this plan
        :param key: the index of the roi
        :return: physician roi or 'uncategorized'
        :rtype: str
        """
        roi_name = self.get_roi_name(key)
        if self.database_rois.is_roi(roi_name):
            if self.database_rois.is_physician(self.physician):
                return self.database_rois.get_physician_roi(self.physician, roi_name)
        return 'uncategorized'

    def get_institutional_roi(self, key):
        """
        Look up the institutional roi of the specified roi for this plan
        :param key: the index of the roi
        :return: institutional roi or 'uncategorized'
        :rtype: str
        """
        roi_name = self.get_roi_name(key)
        if self.database_rois.is_roi(roi_name):
            if self.database_rois.is_physician(self.physician):
                return self.database_rois.get_institutional_roi(self.physician, self.get_physician_roi(key))
            if roi_name in self.database_rois.institutional_rois:
                return roi_name
        return 'uncategorized'

    @property
    def ptv_exists(self):
        """
        Check if the plan has at least one PTV assigned
        :rtype: bool
        """
        if self.dicompyler_rt_structures:
            for key in list(self.dicompyler_rt_structures):
                if self.get_roi_type(key) == 'PTV':
                    return True
        return False

    @property
    def roi_names(self):
        """
        :return: roi_names for this plan per the plan (clean_name not applied)
        :rtype: list
        """
        return [self.get_roi_name(key) for key in list(self.dicompyler_rt_structures)]

    def get_roi_key(self, roi_name):
        """
        Look-up the roi index for a given roi_name
        :param roi_name: the ROI name to be looked-up
        :return: the roi index or None
        :rtype: int
        """
        roi_name = clean_name(roi_name)
        for key in list(self.dicompyler_rt_structures):
            if roi_name == clean_name(self.get_roi_name(key)):
                return key

    @property
    def ptv_names(self):
        return self.stored_values['ptv_names']

    @property
    def init_param(self):
        params = ['plan_file', 'structure_file', 'dose_file', 'plan_ptvs',
                  'plan_over_rides', 'global_plan_over_rides', 'roi_over_ride']
        return {key: getattr(self, key) for key in params}

    @property
    def validation(self):
        """
        :return: data used in GUI to indicate validity of data read-in from DICOM files to attract user to edit data
        :rtype: dict
        """
        return {'physician': {'status': self.database_rois.is_physician(self.physician) and self.physician != 'DEFAULT',
                              'value': self.physician,
                              'message': "No physician assigned or physician is not in ROI Map."},
                'mrn': {'status': self.mrn is not None,
                        'value': self.mrn,
                        'message': "MRN is empty."},
                'study_instance_uid': {'status': self.is_study_instance_uid_to_be_imported_valid,
                                       'value': self.study_instance_uid_to_be_imported,
                                       'message': "Study Instance UID already exists in the database."},
                'ptv': {'status': self.ptv_exists,
                        'value': self.ptv_names,
                        'message': "No PTV found."},
                'sim_study_date': {'status': is_date(self.sim_study_date),
                                   'value': self.sim_study_date,
                                   'message': "Simulation date is empty or invalid."},
                'rx_dose': {'status': self.rx_dose is not None,
                            'value': self.rx_dose,
                            'message': "Prescription dose is not defined."},
                'complete_file_set': {'status': self.file_set_complete,
                                      'value': self.file_set_complete,
                                      'message': ["Missing RT %s." % ', '.join(self.missing_files), '']
                                      [self.file_set_complete]}}

    @property
    def warning(self):
        msg = ''
        incomplete = False
        if self.plan_file and self.dose_file and self.structure_file:
            validation = self.validation
            failed_keys = {key for key, value in validation.items() if not value['status']}
            if failed_keys:
                if 'complete_file_set' in failed_keys:
                    msg = "ERROR: %s" % validation['complete_file_set']['message']
                    incomplete = True
                else:
                    msg = "WARNING: %s" % ' '.join([validation[key]['message'] for key in failed_keys])
        else:
            msg = "ERROR: Incomplete Fileset. RT Plan, Dose, and Structure required."
            incomplete = True

        return {'label': msg, 'incomplete': incomplete}

    def autodetect_target_roi_type(self, key=None):
        """
        Target/tumor ROIs are often not labeled properly in ROI Type, but often start with correct acronym
        any ROI that starts with GTV, CTV, ITV, or PTV and is followed by nothing, a single number, or a character
        then a single number will be flagged.
        :param key: the index of the roi, if key is None, will search all keys
        """

        if key is None:
            roi_names = {key: structure['name'].lower() for key, structure in self.dicompyler_rt_structures.items()}
        else:
            roi_names = {key: self.dicompyler_rt_structures[key]['name'].lower()}

        for key, roi_name in roi_names.items():
            roi_name_len = len(roi_name)
            targets = {'gtv', 'ctv', 'itv', 'ptv'}
            if self.get_physician_roi(key).lower() in targets:
                self.roi_over_ride['type'][key] = self.get_physician_roi(key).upper()
            elif (roi_name_len > 2 and roi_name[0:3] in targets) and \
                    ((roi_name_len == 3) or
                     (roi_name_len == 4 and roi_name[3].isdigit()) or
                     (roi_name_len == 5 and not roi_name[3].isdigit() and roi_name[4].isdigit())):
                self.roi_over_ride['type'][key] = roi_name[0:3].upper()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# options.py
"""
Class used to manage user options
"""
# Copyright (c) 2016-2019 Dan Cutright
# This file is part of DVH Analytics, released under a BSD license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVH-Analytics

import pickle
from os.path import isfile
from os import unlink
import hashlib
from copy import deepcopy
from dvha._version import __version__
from dvha.paths import OPTIONS_PATH, OPTIONS_CHECKSUM_PATH, INBOX_DIR, IMPORTED_DIR, REVIEW_DIR


class DefaultOptions:
    """Create default options, to be inherited by Options class"""
    def __init__(self):
        self.VERSION = __version__

        self.DB_TYPE = 'sqlite'
        self.SQL_PGSQL_IP_HIST = []
        self.DEFAULT_CNF = {'pgsql': {'host': 'localhost',
                                      'dbname': 'dvh',
                                      'port': '5432'},
                            'sqlite': {'host': 'dvha.db'}}
        self.SQL_LAST_CNX = deepcopy(self.DEFAULT_CNF)
        self._sql_vars = ['DB_TYPE', 'SQL_PGSQL_IP_HIST', 'DEFAULT_CNF', 'SQL_LAST_CNX']

        self.MIN_BORDER = 50

        self.MAX_FIELD_SIZE_X = 400  # in mm
        self.MAX_FIELD_SIZE_Y = 400  # in mm

        # These colors propagate to all tabs that visualize your two groups
        self.PLOT_COLOR = 'blue'
        self.PLOT_COLOR_2 = 'red'

        # The line width and style of selected DVHs in the DVH plot
        self.DVH_LINE_WIDTH = 2
        self.DVH_LINE_DASH = 'solid'

        # Adjusts the opacity of the inner-quartile ranges
        self.IQR_ALPHA = 0.1

        # Adjust the plot font sizes
        self.PLOT_AXIS_LABEL_FONT_SIZE = "12pt"
        self.PLOT_AXIS_MAJOR_LABEL_FONT_SIZE = "10pt"

        # Number of data points are reduced by this factor during dynamic plot interaction to speed-up visualizations
        # This is only applied to the DVH plot since it has a large amount of data
        self.LOD_FACTOR = 100

        # All DVHs in SQL DB have 1cGy bin widths regardless of this value.  However, the queried DVHs will be
        # down-sampled using this bin_width
        self.dvh_bin_width = 5

        # Options for the group statistical DVHs in the DVHs tab
        self.STATS_MEDIAN_LINE_WIDTH = 1
        self.STATS_MEDIAN_LINE_DASH = 'solid'
        self.STATS_MEDIAN_ALPHA = 0.6
        self.STATS_MEAN_LINE_WIDTH = 2
        self.STATS_MEAN_LINE_DASH = 'dashed'
        self.STATS_MEAN_ALPHA = 0.5
        self.STATS_MAX_LINE_WIDTH = 1
        self.STATS_MAX_LINE_DASH = 'dotted'
        self.STATS_MAX_ALPHA = 1
        self.STATS_MIN_LINE_WIDTH = 1
        self.STATS_MIN_LINE_DASH = 'dotted'
        self.STATS_MIN_ALPHA = 1

        # Options for the time-series plot
        self.CORRELATION_POS_COLOR_1 = 'blue'
        self.CORRELATION_NEG_COLOR_1 = 'green'
        self.CORRELATION_POS_COLOR_2 = 'red'
        self.CORRELATION_NEG_COLOR_2 = 'purple'
        self.CORRELATION_MATRIX_VARS = ['Beam Area (Mean)', 'Beam Dose (Mean)', 'Beam MU (Mean)',
                                        'Beam Perimeter (Mean)', 'PTV Cross-Section Median', 'PTV Distance (Centroids)',
                                        'PTV Distance (Max)', 'PTV Distance (Mean)', 'PTV Distance (Median)',
                                        'PTV Distance (Min)', 'PTV Max Dose', 'PTV Min Dose', 'PTV Surface Area',
                                        'PTV Volume', 'Plan Complexity', 'ROI Cross-Section Max',
                                        'ROI Cross-Section Median', 'ROI Max Dose', 'ROI Mean Dose', 'ROI Min Dose',
                                        'ROI Surface Area', 'ROI Volume', 'Rx Dose', 'Total Plan MU']

        # Options for the time-series plot
        self.TIME_SERIES_CIRCLE_SIZE = 10
        self.TIME_SERIES_CIRCLE_ALPHA = 0.3
        self.TIME_SERIES_TREND_LINE_WIDTH = 1
        self.TIME_SERIES_TREND_LINE_DASH = 'solid'
        self.TIME_SERIES_AVG_LINE_WIDTH = 1
        self.TIME_SERIES_AVG_LINE_DASH = 'dotted'
        self.TIME_SERIES_PATCH_ALPHA = 0.1

        # Options for the time-series plot
        self.CONTROL_CHART_CIRCLE_SIZE = 10
        self.CONTROL_CHART_CIRCLE_ALPHA = 0.3
        self.CONTROL_CHART_LINE_WIDTH = 1
        self.CONTROL_CHART_LINE_DASH = 'solid'
        self.CONTROL_CHART_LINE_COLOR = 'black'
        self.CONTROL_CHART_CENTER_LINE_WIDTH = 2
        self.CONTROL_CHART_CENTER_LINE_DASH = 'solid'
        self.CONTROL_CHART_CENTER_LINE_COLOR = 'black'
        self.CONTROL_CHART_CENTER_LINE_ALPHA = 1
        self.CONTROL_CHART_UCL_LINE_WIDTH = 2
        self.CONTROL_CHART_UCL_LINE_DASH = 'dashed'
        self.CONTROL_CHART_UCL_LINE_COLOR = 'red'
        self.CONTROL_CHART_UCL_LINE_ALPHA = 1
        self.CONTROL_CHART_LCL_LINE_WIDTH = 2
        self.CONTROL_CHART_LCL_LINE_DASH = 'dashed'
        self.CONTROL_CHART_LCL_LINE_COLOR = 'red'
        self.CONTROL_CHART_LCL_LINE_ALPHA = 1
        self.CONTROL_CHART_PATCH_ALPHA = 0.1
        self.CONTROL_CHART_PATCH_COLOR = 'grey'
        self.CONTROL_CHART_OUT_OF_CONTROL_COLOR = 'green'
        self.CONTROL_CHART_OUT_OF_CONTROL_COLOR_2 = 'purple'
        self.CONTROL_CHART_OUT_OF_CONTROL_ALPHA = 1

        # Adjust the opacity of the histograms
        self.HISTOGRAM_ALPHA = 0.3

        # Options for the plot in the Multi-Variable Regression tab
        self.REGRESSION_CIRCLE_SIZE = 10
        self.REGRESSION_ALPHA = 0.5
        self.REGRESSION_LINE_WIDTH = 2
        self.REGRESSION_LINE_DASH = 'dashed'

        self.REGRESSION_RESIDUAL_CIRCLE_SIZE = 3
        self.REGRESSION_RESIDUAL_ALPHA = 0.5
        self.REGRESSION_RESIDUAL_LINE_WIDTH = 2
        self.REGRESSION_RESIDUAL_LINE_DASH = 'solid'
        self.REGRESSION_RESIDUAL_LINE_COLOR = 'black'

        # Random forest
        self.MACHINE_LEARNING_ALPHA = 0.5
        self.MACHINE_LEARNING_ALPHA_DIFF = 0.35
        self.MACHINE_LEARNING_SIZE_PREDICT = 5
        self.MACHINE_LEARNING_SIZE_DATA = 5
        self.MACHINE_LEARNING_SIZE_MULTI_VAR = 5
        self.MACHINE_LEARNING_COLOR_PREDICT = 'blue'
        self.MACHINE_LEARNING_COLOR_DATA = 'black'
        self.MACHINE_LEARNING_COLOR_MULTI_VAR = 'red'

        # This is the number of bins up do 100% used when resampling a DVH to fractional dose
        self.RESAMPLED_DVH_BIN_COUNT = 5000

        self.COMPLEXITY_SCORE_X_WEIGHT = 1.
        self.COMPLEXITY_SCORE_Y_WEIGHT = 1.

        # Per TG-263 (plus ITV)
        self.ROI_TYPES = ['ORGAN', 'PTV', 'ITV', 'CTV', 'GTV',
                          'AVOIDANCE', 'BOLUS', 'CAVITY', 'CONTRAST_AGENT', 'EXTERNAL',
                          'IRRAD_VOLUME', 'REGISTRATION', 'TREATED_VOLUME']

        self.KEEP_IN_INBOX = 0
        self.SEARCH_SUBFOLDERS = 1
        self.IMPORT_UNCATEGORIZED = 0

        self.INBOX_DIR = INBOX_DIR
        self.IMPORTED_DIR = IMPORTED_DIR
        self.REVIEW_DIR = REVIEW_DIR

        self.MAX_DOSE_VOLUME = 0.03

        self.USE_DICOM_DVH = False
        self.AUTO_SUM_DOSE = True

        self.save_fig_param = {'y_range_start': -0.0005,
                               'x_range_start': 0,
                               'y_range_end': 1.0005,
                               'background_fill_color': 'none',
                               'border_fill_color': 'none',
                               'plot_height': 600,
                               'plot_width': 820}


class Options(DefaultOptions):
    def __init__(self):
        DefaultOptions.__init__(self)
        self.__set_option_attr()

        self.load()

    def __set_option_attr(self):
        option_attr = []
        for attr in self.__dict__:
            if not attr.startswith('_'):
                option_attr.append(attr)
        self.option_attr = option_attr

    def load(self):

        if isfile(OPTIONS_PATH) and self.validate_options_file():
            try:
                with open(OPTIONS_PATH, 'rb') as infile:
                    loaded_options = pickle.load(infile)
            except EOFError:
                print('ERROR: Options file corrupted. Loading default options.')
                loaded_options = {}

            for key, value in loaded_options.items():
                if hasattr(self, key):
                    setattr(self, key, value)

    def save(self):

        out_options = {}
        for attr in self.option_attr:
            out_options[attr] = getattr(self, attr)
        out_options['VERSION'] = DefaultOptions().VERSION
        with open(OPTIONS_PATH, 'wb') as outfile:
            pickle.dump(out_options, outfile)
        self.save_checksum()

    def set_option(self, attr, value):
        """
        Change or create an option value
        :param attr: name of option
        :type attr: str
        :param value: value of option
        """
        if not hasattr(self, attr):
            print('WARNING: This option did not previously exist!')
        setattr(self, attr, value)

    def save_checksum(self):
        check_sum = self.calculate_checksum()
        if check_sum:
            with open(OPTIONS_CHECKSUM_PATH, 'w') as outfile:
                outfile.write(check_sum)

    @staticmethod
    def calculate_checksum():
        if isfile(OPTIONS_PATH):
            with open(OPTIONS_PATH, 'rb') as infile:
                options_str = str(infile.read())
            return hashlib.md5(options_str.encode('utf-8')).hexdigest()
        return None

    @staticmethod
    def load_stored_checksum():
        if isfile(OPTIONS_CHECKSUM_PATH):
            with open(OPTIONS_CHECKSUM_PATH, 'r') as infile:
                checksum = infile.read()
            return checksum
        return None

    def validate_options_file(self):
        try:
            current_checksum = self.calculate_checksum()
            stored_checksum = self.load_stored_checksum()
            if current_checksum == stored_checksum:
                return True
        except Exception:
            print('Corrupted options file detected. Loading default options.')
            return False

    def restore_defaults(self):
        if isfile(OPTIONS_PATH):
            unlink(OPTIONS_PATH)
        if isfile(OPTIONS_CHECKSUM_PATH):
            unlink(OPTIONS_CHECKSUM_PATH)
        default_options = DefaultOptions()

        for attr in default_options.__dict__:
            if not attr.startswith('_') and attr not in self._sql_vars:
                setattr(self, attr, getattr(default_options, attr))




#!/usr/bin/env python
# -*- coding: utf-8 -*-

# dialogs.protocols.py
"""
Dialogs used to edit DVH Constraint Protocols
"""
# Copyright (c) 2016-2019 Dan Cutright
# This file is part of DVH Analytics, released under a BSD license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVH-Analytics

import wx
from dvha.models.data_table import DataTable
from dvha.models.dvh import Protocols, Constraint
from dvha.tools.roi_name_manager import clean_name
from dvha.tools.utilities import get_window_size, get_selected_listctrl_items


class ProtocolsEditor(wx.Dialog):
    def __init__(self, roi_map):
        wx.Dialog.__init__(self, None)

        self.roi_map = roi_map
        self.protocols = Protocols()

        keys = ['protocol', 'fractionation']
        self.combo_box = {key: wx.ComboBox(self, wx.ID_ANY, choices=[], style=wx.CB_DROPDOWN | wx.CB_READONLY)
                          for key in keys}

        button_map = {'add': '+', 'del': '-', 'edit': u"Δ"}
        self.buttons_combos = {key: {b_key: wx.Button(self, wx.ID_ANY, b_val)
                                     for b_key, b_val in button_map.items()}
                               for key in keys}

        keys = ['Add', 'Edit', 'Delete', 'Select All', 'Deselect All']
        self.button_constraints = {key: wx.Button(self, wx.ID_ANY, key) for key in keys}

        button_id_map = {'Save': wx.ID_ANY, 'OK': wx.ID_OK, 'Cancel': wx.ID_CANCEL}
        self.button_dlg = {key: wx.Button(self, wxid, key) for key, wxid in button_id_map.items()}

        self.list_ctrl = wx.ListCtrl(self, wx.ID_ANY, style=wx.LC_HRULES | wx.LC_REPORT | wx.LC_VRULES)
        self.data_table = DataTable(self.list_ctrl, widths=[-2, -2, -2])

        self.__set_properties()
        self.__do_bind()
        self.__do_layout()

        self.update_table()

        self.ShowModal()
        self.Destroy()

    def __set_properties(self):
        self.SetMinSize(get_window_size(0.4, 0.7))
        self.SetTitle("Protocol Editor")

        self.update_protocols()

        for key in ['protocol', 'fractionation']:
            for button in self.buttons_combos[key].values():
                button.SetMaxSize((25, 25))

        self.button_constraints['Edit'].Disable()
        self.button_constraints['Delete'].Disable()

        self.combo_box['protocol'].SetMinSize((300, self.combo_box['protocol'].GetSize()[1]))
        self.combo_box['fractionation'].SetMinSize((50, self.combo_box['fractionation'].GetSize()[1]))

    def __do_bind(self):
        self.Bind(wx.EVT_COMBOBOX, self.protocol_ticker, id=self.combo_box['protocol'].GetId())
        self.Bind(wx.EVT_COMBOBOX, self.update_table, id=self.combo_box['fractionation'].GetId())
        self.Bind(wx.EVT_BUTTON, self.on_select_all, id=self.button_constraints['Select All'].GetId())
        self.Bind(wx.EVT_BUTTON, self.on_deselect_all, id=self.button_constraints['Deselect All'].GetId())
        self.Bind(wx.EVT_BUTTON, self.on_delete_constraint, id=self.button_constraints['Delete'].GetId())
        self.Bind(wx.EVT_LIST_ITEM_SELECTED, self.update_delete_constraint_enable, id=self.list_ctrl.GetId())
        self.Bind(wx.EVT_LIST_ITEM_DESELECTED, self.update_delete_constraint_enable, id=self.list_ctrl.GetId())
        self.Bind(wx.EVT_BUTTON, self.on_delete_protocol, id=self.buttons_combos['protocol']['del'].GetId())
        self.Bind(wx.EVT_BUTTON, self.on_delete_fractionation, id=self.buttons_combos['fractionation']['del'].GetId())
        self.Bind(wx.EVT_BUTTON, self.on_add_constraint, id=self.button_constraints['Add'].GetId())

    def __do_layout(self):

        sizer_wrapper = wx.BoxSizer(wx.VERTICAL)
        sizer_main = wx.BoxSizer(wx.VERTICAL)
        sizer_combos = {key: wx.StaticBoxSizer(wx.StaticBox(self, wx.ID_ANY, key.capitalize()), wx.HORIZONTAL)
                        for key in ['protocol', 'fractionation']}
        sizer_combos_wrapper = wx.BoxSizer(wx.HORIZONTAL)
        sizer_constraints = wx.StaticBoxSizer(wx.StaticBox(self, wx.ID_ANY, 'Constraints'), wx.VERTICAL)
        sizer_actions = wx.BoxSizer(wx.HORIZONTAL)
        sizer_dlg_buttons = wx.BoxSizer(wx.HORIZONTAL)

        for key in ['protocol', 'fractionation']:
            sizer_combos[key].Add(self.combo_box[key], 0, wx.LEFT | wx.RIGHT, 2)
            sizer_combos[key].Add((15, 10), 0, 0, 0)
            sizer_combos[key].Add(self.buttons_combos[key]['add'], 0, wx.LEFT | wx.RIGHT, 2)
            sizer_combos[key].Add(self.buttons_combos[key]['del'], 0, wx.LEFT | wx.RIGHT, 2)
            sizer_combos[key].Add(self.buttons_combos[key]['edit'], 0, wx.LEFT, 2)
            sizer_combos_wrapper.Add(sizer_combos[key], 0, wx.ALL | wx.EXPAND, 5)
        sizer_main.Add(sizer_combos_wrapper, 0, wx.EXPAND, 0)

        sizer_actions.Add(self.button_constraints['Add'], 0, wx.RIGHT, 5)
        sizer_actions.Add(self.button_constraints['Edit'], 0, wx.RIGHT | wx.LEFT, 5)
        sizer_actions.Add(self.button_constraints['Delete'], 0, wx.RIGHT | wx.LEFT, 5)
        sizer_actions.Add(self.button_constraints['Select All'], 0, wx.RIGHT | wx.LEFT, 5)
        sizer_actions.Add(self.button_constraints['Deselect All'], 0,  wx.LEFT, 5)

        sizer_constraints.Add(sizer_actions, 0, wx.ALL | wx.EXPAND, 5)
        sizer_constraints.Add(self.list_ctrl, 1, wx.ALL | wx.EXPAND, 5)
        sizer_main.Add(sizer_constraints, 1, wx.ALL | wx.EXPAND, 5)

        sizer_dlg_buttons.Add(self.button_dlg['Save'], 0, wx.RIGHT, 5)
        sizer_dlg_buttons.Add(self.button_dlg['OK'], 0, wx.LEFT | wx.RIGHT, 5)
        sizer_dlg_buttons.Add(self.button_dlg['Cancel'], 0, wx.LEFT | wx.RIGHT, 5)
        sizer_main.Add(sizer_dlg_buttons, 0, wx.ALIGN_RIGHT | wx.ALL, 5)

        sizer_wrapper.Add(sizer_main, 1, wx.ALL | wx.EXPAND, 10)

        self.SetSizer(sizer_wrapper)
        self.Layout()
        self.Fit()
        self.Center()

    @property
    def selected_protocol(self):
        return self.combo_box['protocol'].GetValue()

    @property
    def selected_fx(self):
        return self.combo_box['fractionation'].GetValue()

    def update_table(self, *evt):
        if self.selected_protocol and self.selected_fx:
            data, columns = self.protocols.get_column_data(self.selected_protocol, self.selected_fx)
            self.data_table.set_data(data, columns)

    def update_protocols(self, set_name=None):
        protocol_names = self.protocols.protocol_names
        if not protocol_names:
            self.clear_data()
        else:
            self.combo_box['protocol'].SetItems(protocol_names)
            selected_protocol = set_name if set_name is not None else protocol_names[0]
            self.combo_box['protocol'].SetValue(selected_protocol)
            self.update_fractionations()

    def update_fractionations(self):
        old_fx = self.selected_fx
        fxs = self.protocols.get_fractionations(self.selected_protocol)
        self.combo_box['fractionation'].SetItems(fxs)
        if not fxs:
            self.data_table.clear()
        else:
            new_fx = fxs[0] if old_fx not in fxs else old_fx
            self.combo_box['fractionation'].SetValue(new_fx)

    def protocol_ticker(self, *evt):
        self.update_fractionations()
        self.update_table()
        self.update_del_combos_enable()

    def on_select_all(self, *evt):
        self.data_table.apply_selection_to_all(True)

    def on_deselect_all(self, *evt):
        self.data_table.apply_selection_to_all(False)

    def on_delete_protocol(self, *evt):
        self.protocols.delete_protocol(self.selected_protocol)
        self.update_protocols()
        self.update_table()
        self.update_del_combos_enable()

    def on_delete_fractionation(self, *evt):
        self.protocols.delete_fractionation(self.selected_protocol, self.selected_fx)
        self.update_fractionations()
        self.update_table()
        self.update_del_combos_enable()

    def update_del_combos_enable(self):
        self.buttons_combos['protocol']['del'].Enable(bool(self.selected_protocol))
        self.buttons_combos['fractionation']['del'].Enable(bool(self.selected_fx))

    def on_delete_constraint(self, *evt):
        for index_and_row in self.data_table.selected_row_data_with_index:
            index, row = tuple(index_and_row)

            # the data table view does not print repeated structure names,
            # if roi is blank, work backwards until it isn't
            i = index
            while not self.data_table.data['Structure'][i] and i != -1:  # prevent infinite loop if no roi found
                i -= 1
            roi = self.data_table.data['Structure'][i]

            constraint_label = row[2].replace('cc', '').replace('Gy', '')
            if roi:
                self.protocols.delete_constraint(self.selected_protocol, self.selected_fx, roi, constraint_label)
        self.update_table()

    def on_add_constraint(self, *evt):
        existing_rois = self.protocols.get_rois(self.selected_protocol, self.selected_fx)
        dlg = AddProtocolROI(self, existing_rois)
        res = dlg.ShowModal()
        if res == wx.ID_OK:
            EditProtocolROI(self, dlg.roi_name, self.protocols, self.selected_protocol, self.selected_fx, self.roi_map)

    def update_delete_constraint_enable(self, *evt):
        selected = get_selected_listctrl_items(self.list_ctrl)
        count = len([i for i in selected if self.data_table.get_row(i)[-1]])
        self.button_constraints['Delete'].Enable(count > 0)
        self.button_constraints['Edit'].Enable(count == 1)

    def clear_data(self):
        for combo_box in self.combo_box.values():
            combo_box.SetItems([''])
        self.data_table.clear()


class AddProtocolROI(wx.Dialog):
    def __init__(self, parent, existing_rois):
        wx.Dialog.__init__(self, parent)
        self.existing_rois = existing_rois
        self.text_ctrl = wx.TextCtrl(self, wx.ID_ANY, "")
        self.button_ok = wx.Button(self, wx.ID_OK, "OK")
        self.button_cancel = wx.Button(self, wx.ID_CANCEL, "Cancel")

        self.__set_properties()
        self.__do_bind()
        self.__do_layout()

    def __set_properties(self):
        self.SetTitle('Add New ROI to Protocol')
        self.SetMinSize((400, 0))

    def __do_bind(self):
        self.Bind(wx.EVT_TEXT, self.update_ok_enable, id=self.text_ctrl.GetId())

    def __do_layout(self):
        sizer_wrapper = wx.BoxSizer(wx.VERTICAL)
        sizer_main = wx.BoxSizer(wx.VERTICAL)
        sizer_input = wx.StaticBoxSizer(wx.StaticBox(self, wx.ID_ANY, "New ROI Name"), wx.VERTICAL)
        sizer_buttons = wx.BoxSizer(wx.HORIZONTAL)

        sizer_input.Add(self.text_ctrl, 0, wx.EXPAND | wx.ALL, 5)
        sizer_main.Add(sizer_input, 0, wx.EXPAND | wx.ALL, 5)

        sizer_buttons.Add(self.button_ok, 0, wx.EXPAND | wx.ALL, 5)
        sizer_buttons.Add(self.button_cancel, 0, wx.EXPAND | wx.ALL, 5)
        sizer_main.Add(sizer_buttons, 0, wx.ALIGN_RIGHT | wx.ALL, 5)

        sizer_wrapper.Add(sizer_main, 0, wx.EXPAND | wx.ALL, 5)

        self.SetSizer(sizer_wrapper)
        self.Layout()
        self.Fit()
        self.Center()

    @property
    def roi_name(self):
        return self.text_ctrl.GetValue()

    @property
    def is_roi_valid(self):
        clean_roi = clean_name(self.roi_name)
        for roi in self.existing_rois:
            if clean_roi == clean_name(roi):
                return False
        return True

    def update_ok_enable(self, *evt):
        self.button_ok.Enable(self.is_roi_valid)


class EditProtocolROI(wx.Dialog):
    def __init__(self, parent, roi_name, protocols, protocol_name, fractionation, roi_map):
        wx.Dialog.__init__(self, parent)

        self.parent = parent
        self.protocols = protocols
        self.protocol_name = protocol_name
        self.fractionation = fractionation
        self.roi_map = roi_map

        # columns = ['Input Type', 'Input Value', 'Operator', 'Output Type', 'Output Value']
        keys = ['constraints', 'aliases']
        style = wx.LC_HRULES | wx.LC_REPORT | wx.LC_VRULES
        self.list_ctrl = {key: wx.ListCtrl(self, wx.ID_ANY, style=style) for key in keys}
        self.data_table = {key: DataTable(self.list_ctrl[key], widths=[-2]) for i, key in enumerate(keys)}

        key_map = {'Add': '+', 'Delete': '-', 'Help': 'Help'}
        self.button_constraint = {key: wx.Button(self, wx.ID_ANY, label) for key, label in key_map.items()}
        self.button_alias = {key: wx.Button(self, wx.ID_ANY, key) for key in ['Add', 'Remove']}
        self.button_ok = wx.Button(self, wx.ID_OK, "OK")
        self.button_cancel = wx.Button(self, wx.ID_CANCEL, "Cancel")

        self.combo_box_alias = wx.ComboBox(self, wx.ID_ANY, choices=[], style=wx.CB_DROPDOWN | wx.CB_READONLY)

        self.text_ctrl_structure = wx.TextCtrl(self, wx.ID_ANY, roi_name)
        self.text_ctrl_structure.Disable()

        self.text_ctrl_constraint = wx.TextCtrl(self, wx.ID_ANY, "")

        self.__set_properties()
        self.__do_bind()
        self.__do_layout()

        self.ShowModal()
        self.Destroy()

    def __set_properties(self):
        self.SetTitle("Constraints Editor")
        self.update_combo_box_aliases()

        for key in ['Add', 'Delete']:
            self.button_constraint[key].SetMaxSize((25, 25))

        self.text_ctrl_constraint.SetMinSize((200, self.text_ctrl_constraint.GetSize()[1]))

        self.button_constraint['Add'].Disable()
        self.button_constraint['Delete'].Disable()

    def __do_bind(self):
        self.Bind(wx.EVT_TEXT, self.update_add_enable, id=self.text_ctrl_constraint.GetId())
        self.Bind(wx.EVT_BUTTON, self.on_add_constraint, id=self.button_constraint['Add'].GetId())

    def __do_layout(self):
        sizer_wrapper = wx.BoxSizer(wx.VERTICAL)
        sizer_main = wx.BoxSizer(wx.VERTICAL)
        sizer_structure = wx.StaticBoxSizer(wx.StaticBox(self, wx.ID_ANY, "Structure Name"), wx.VERTICAL)
        sizer_constraints_wrapper = wx.StaticBoxSizer(wx.StaticBox(self, wx.ID_ANY, "Constraints"), wx.VERTICAL)
        sizer_constraints = wx.BoxSizer(wx.HORIZONTAL)
        sizer_aliases_wrapper = wx.StaticBoxSizer(wx.StaticBox(self, wx.ID_ANY, "Aliases"), wx.VERTICAL)
        sizer_aliases = wx.BoxSizer(wx.VERTICAL)
        sizer_aliases_edit = wx.BoxSizer(wx.HORIZONTAL)
        sizer_ok_cancel = wx.BoxSizer(wx.HORIZONTAL)

        label = wx.StaticText(self, wx.ID_ANY, "Protocol: %s %sFx" % (self.protocol_name, self.fractionation))
        sizer_main.Add(label, 0, wx.ALL, 5)

        sizer_structure.Add(self.text_ctrl_structure, 0, wx.EXPAND | wx.ALL, 5)
        sizer_main.Add(sizer_structure, 0, wx.ALL | wx.EXPAND, 5)

        sizer_constraints.Add(self.text_ctrl_constraint, 0, wx.RIGHT, 5)
        sizer_constraints.Add(self.button_constraint['Add'], 0, wx.LEFT | wx.RIGHT, 5)
        sizer_constraints.Add(self.button_constraint['Delete'], 0, wx.LEFT | wx.RIGHT, 5)
        sizer_constraints.Add(self.button_constraint['Help'], 0, wx.LEFT, 5)
        sizer_constraints_wrapper.Add(sizer_constraints, 0, wx.ALL, 5)
        sizer_constraints_wrapper.Add(self.list_ctrl['constraints'], 1, wx.ALL | wx.EXPAND, 5)
        sizer_main.Add(sizer_constraints_wrapper, 1, wx.ALL | wx.EXPAND, 5)

        label_alias = wx.StaticText(self, wx.ID_ANY, "Institutional ROI:")
        sizer_aliases.Add(label_alias, 0, 0, 0)
        sizer_aliases.Add(sizer_aliases_edit, 0, wx.ALL | wx.EXPAND, 0)
        sizer_aliases_wrapper.Add(sizer_aliases, 0, wx.ALL | wx.EXPAND, 5)
        sizer_main.Add(sizer_aliases_wrapper, 1, wx.EXPAND, 0)

        sizer_aliases_edit.Add(self.combo_box_alias, 1, wx.RIGHT, 5)
        sizer_aliases_edit.Add(self.button_alias['Add'], 0, wx.LEFT | wx.RIGHT, 5)
        sizer_aliases_edit.Add(self.button_alias['Remove'], 0, wx.LEFT, 5)
        sizer_aliases_wrapper.Add(self.list_ctrl['aliases'], 1, wx.ALL | wx.EXPAND, 5)

        sizer_ok_cancel.Add(self.button_ok, 0, wx.ALL, 5)
        sizer_ok_cancel.Add(self.button_cancel, 0, wx.ALL, 5)
        sizer_main.Add(sizer_ok_cancel, 0, wx.ALIGN_RIGHT | wx.ALL, 5)

        sizer_wrapper.Add(sizer_main, 1, wx.ALL | wx.EXPAND, 5)

        self.SetSizer(sizer_wrapper)
        self.Layout()
        self.Fit()
        self.Center()

    def update_combo_box_aliases(self, set_value=None):
        self.combo_box_alias.SetItems(self.roi_map.institutional_rois)
        if set_value is not None and set_value in self.roi_map.institutional_rois:
            self.combo_box_alias.SetValue(set_value)
        else:
            self.combo_box_alias.SetValue(self.roi_map.institutional_rois[0])

    def update_add_enable(self, *evt):
        self.button_constraint['Add'].Enable(self.constraint is not None)

    @property
    def constraint(self):
        """Given a string, convert into constraint name, operator, and threshold"""
        text = self.text_ctrl_constraint.GetValue()
        # include one and only one < or > operator, no =
        iter_ = [c in text for c in ['<', '>']]
        if any(iter_) and not all(iter_) and '=' not in text:
            if text.count('<') == 1 or text.count('>') == 1:
                operator = ['<', '>']['>' in text]
                name, threshold = tuple([t.strip().upper() for t in text.split(operator)])
                if '_' in name and (name[0] != '_' and name[-1] != '_'):
                    if any([c == name.split('_')[0] for c in ['V', 'D', 'MVS']]):
                        return Constraint(name, operator, threshold)

    @property
    def structure(self):
        return self.text_ctrl_structure.GetValue().strip()

    def on_add_constraint(self, *evt):
        self.protocols.add_constraint(self.protocol_name, self.fractionation, self.structure, self.constraint)
        self.update_constraints_table()

    def update_constraints_table(self, *evt):
        constraints = self.protocols.get_constraints(self.protocol_name, self.fractionation, self.structure)
        constraints = {'Constraint': [str(c) for c in constraints]}
        # data, columns = self.protocols.get_column_data(self.protocol_name, self.fractionation, roi=self.structure)
        self.data_table['constraints'].set_data(constraints, list(constraints))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# models.control_chart.py
"""
Class for the Control Chart frame in the main view
"""
# Copyright (c) 2016-2019 Dan Cutright
# This file is part of DVH Analytics, released under a BSD license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVH-Analytics

import wx
from pubsub import pub
from copy import deepcopy
from dvha.models.plot import PlotControlChart
from dvha.db import sql_columns
from dvha.dialogs.export import save_data_to_file


class ControlChartFrame:
    """
    Object to be passed into notebook panel for the Control Chart tab
    """
    def __init__(self, parent, group_data, options):
        """
        :param parent:  notebook panel in main view
        :type parent: Panel
        :param group_data: dvh, table, and stats data
        :type group_data: dict
        :param options: user options containing visual preferences
        :type options: Options
        """
        self.parent = parent
        self.group_data = group_data
        self.choices = []
        self.models = {grp: {} for grp in [1, 2]}

        self.group = 1

        self.y_axis_options = sql_columns.numerical

        self.combo_box_y_axis = wx.ComboBox(self.parent, wx.ID_ANY, style=wx.CB_DROPDOWN | wx.CB_READONLY)

        self.button_export = wx.Button(self.parent, wx.ID_ANY, "Export")
        self.button_save_plot = wx.Button(self.parent, wx.ID_ANY, "Save Plot")
        self.plot = PlotControlChart(self.parent, options)

        self.__do_bind()
        self.__set_properties()
        self.__do_subscribe()
        self.__do_layout()

    def __do_bind(self):
        self.parent.Bind(wx.EVT_COMBOBOX, self.on_combo_box_y, id=self.combo_box_y_axis.GetId())
        self.parent.Bind(wx.EVT_BUTTON, self.on_save_plot, id=self.button_save_plot.GetId())
        self.parent.Bind(wx.EVT_BUTTON, self.export_csv, id=self.button_export.GetId())

    def __set_properties(self):
        self.combo_box_y_axis.SetMinSize((300, self.combo_box_y_axis.GetSize()[1]))

    def __do_subscribe(self):
        pub.subscribe(self.set_model, "control_chart_set_model")

    def __do_layout(self):
        sizer_wrapper = wx.BoxSizer(wx.VERTICAL)
        sizer_plot = wx.BoxSizer(wx.HORIZONTAL)
        sizer_widgets = wx.StaticBoxSizer(wx.StaticBox(self.parent, wx.ID_ANY, ""), wx.HORIZONTAL)
        sizer_y_axis = wx.BoxSizer(wx.VERTICAL)

        label_y_axis = wx.StaticText(self.parent, wx.ID_ANY, "Charting Variable:")
        sizer_y_axis.Add(label_y_axis, 0, wx.LEFT, 5)
        sizer_y_axis.Add(self.combo_box_y_axis, 0, wx.ALL | wx.EXPAND, 5)
        sizer_widgets.Add(sizer_y_axis, 0, wx.EXPAND, 0)

        sizer_widgets.Add(self.button_export, 0, wx.ALL | wx.EXPAND, 5)
        sizer_widgets.Add(self.button_save_plot, 0, wx.ALL | wx.EXPAND, 5)
        sizer_wrapper.Add(sizer_widgets, 0, wx.EXPAND | wx.BOTTOM, 5)
        sizer_plot.Add(self.plot.layout, 1, wx.EXPAND, 0)
        sizer_wrapper.Add(sizer_plot, 1, wx.EXPAND, 0)

        self.layout = sizer_wrapper

    def update_combo_box_y_choices(self):
        stats_data = self.group_data[self.group]['stats_data']
        if stats_data:
            self.choices = stats_data.trending_variables
            self.choices.sort()
            self.combo_box_y_axis.SetItems(self.choices)
            self.combo_box_y_axis.SetValue('ROI Max Dose')

    @property
    def y_axis(self):
        return self.combo_box_y_axis.GetValue()

    def on_combo_box_y(self, evt):
        self.update_plot()

    def update_plot_ticker(self, evt):
        self.update_plot()

    def update_plot(self):
        stats_data = self.group_data[self.group]['stats_data']
        dates = stats_data.sim_study_dates
        sort_index = sorted(range(len(dates)), key=lambda k: dates[k])
        dates_sorted = [dates[i] for i in sort_index]

        y_values_sorted = [stats_data.data[self.y_axis]['values'][i] for i in sort_index]
        mrn_sorted = [stats_data.mrns[i] for i in sort_index]
        uid_sorted = [stats_data.uids[i] for i in sort_index]

        # remove data with no date
        if 'None' in dates_sorted:
            final_index = dates_sorted.index('None')
            dates_sorted = dates_sorted[:final_index]
            y_values_sorted = y_values_sorted[:final_index]
            mrn_sorted = mrn_sorted[:final_index]
            uid_sorted = uid_sorted[:final_index]

        x = list(range(1, len(dates_sorted)+1))

        self.plot.group = self.group
        self.plot.update_plot(x, y_values_sorted, mrn_sorted, uid_sorted, dates_sorted, y_axis_label=self.y_axis)

        if self.models[self.group] and self.y_axis in self.models[self.group].keys():
            model_data = self.models[self.group][self.y_axis]
            adj_data = self.plot.get_adjusted_control_chart(stats_data=stats_data, **model_data)
            self.plot.update_adjusted_control_chart(**adj_data)

    def update_data(self, group_data):
        self.group_data = group_data
        self.update_combo_box_y_choices()
        self.update_plot()

    @property
    def variables(self):
        stats_data = self.group_data[self.group]['stats_data']
        return list(stats_data)

    # def update_endpoints_and_radbio(self):
    #     if self.dvhs:
    #         if self.dvhs.endpoints['defs']:
    #             for var in self.dvhs.endpoints['defs']['label']:
    #                 if var not in self.variables:
    #                     self.stats_data[var] = {'units': '',
    #                                             'values': self.dvhs.endpoints['data'][var]}
    #
    #             for var in self.variables:
    #                 if var[0:2] in {'D_', 'V_'}:
    #                     if var not in self.dvhs.endpoints['defs']['label']:
    #                         self.stats_data.pop(var)
    #
    #         if self.dvhs.eud:
    #             self.stats_data['EUD'] = {'units': 'Gy',
    #                                       'values': self.dvhs.eud}
    #         if self.dvhs.ntcp_or_tcp:
    #             self.stats_data['NTCP or TCP'] = {'units': '',
    #                                               'values': self.dvhs.ntcp_or_tcp}

    def initialize_y_axis_options(self):
        for i in range(len(self.choices))[::-1]:
            c = self.choices[i]
            if c[0:2] in {'D_', 'V_'} or c in {'EUD', 'NTCP or TCP'}:
                self.choices.pop(i)
        self.choices.sort()
        self.combo_box_y_axis.SetItems(self.choices)
        self.combo_box_y_axis.SetValue('ROI Max Dose')

    def clear_data(self):
        self.initialize_y_axis_options()

    def get_csv(self, selection=None):
        return self.plot.get_csv()

    def export_csv(self, evt):
        save_data_to_file(self.parent, "Export control chart data to CSV", self.plot.get_csv())

    def on_save_plot(self, evt):
        save_data_to_file(self.parent, 'Save control chart', self.plot.html_str,
                          wildcard="HTML files (*.html)|*.html")

    @property
    def has_data(self):
        return self.combo_box_y_axis.IsEnabled()

    def set_model(self, y_variable, x_variables, regression, group):
        self.models[group][y_variable] = {'y_variable': y_variable,
                                          'x_variables': x_variables,
                                          'regression': regression}
        if self.y_axis == y_variable:
            wx.CallAfter(self.update_plot)

    def delete_model(self, y_variable):
        if y_variable in list(self.models):
            self.models.pop(y_variable)

    def get_save_data(self):
        return {'models': self.models}

    def load_save_data(self, saved_data):
        self.models = deepcopy(saved_data['models'])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# tools.errors.py
"""
Classes for DVHA specific error handling
"""
# Copyright (c) 2016-2019 Dan Cutright
# This file is part of DVH Analytics, released under a BSD license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVH-Analytics

import wx


class SQLError(Exception):
    def __init__(self, error_message, failed_sql_command):
        """
        Custom exception class to catch query and update failures in database editor
        :param error_message: The message to be displayed in the SQLErrorDialog
        :type error_message: str
        :param failed_sql_command: the SQL command that failed
        :type failed_sql_command: str
        """
        try:
            self.message = error_message.split('\n')[0]
        except Exception as e:
            print(e)
            self.message = error_message
        self.sql_command = failed_sql_command

    def __str__(self):
        return self.message


class ROIVariationError(Exception):
    def __init__(self, error_message):
        self.message = error_message

    def __str__(self):
        return self.message


class PlottingMemoryError(Exception):
    def __init__(self, error_message):
        self.message = error_message

    def __str__(self):
        return self.message


class ErrorDialog:
    def __init__(self, parent, message, caption, flags=wx.ICON_ERROR | wx.OK | wx.OK_DEFAULT):
        """
        This class allows error messages to be called with a one-liner else-where
        :param parent: wx parent object
        :param message: error message
        :param caption: error title
        :param flags: flags for wx.MessageDialog
        """
        self.dlg = wx.MessageDialog(parent, message, caption, flags)
        self.dlg.Center()
        self.dlg.ShowModal()
        self.dlg.Destroy()


class SQLErrorDialog(ErrorDialog):
    def __init__(self, parent, dvh_sql_error):
        """
        Error dialog using custom SQLError class
        :param parent: the wx parent object
        :param dvh_sql_error: SQLError exception class
        :type dvh_sql_error: SQLError
        """
        ErrorDialog.__init__(self, parent, str(dvh_sql_error), "SQL Syntax Error")


class ROIVariationErrorDialog(ErrorDialog):
    def __init__(self, parent, roi_variation_error):
        """
        Error dialog using custom ROIVariationError class
        :param parent: the wx parent object
        :param roi_variation_error: ROIVariationError exception class
        :type roi_variation_error: ROIVariationError
        """
        ErrorDialog.__init__(self, parent, str(roi_variation_error), "ROI Variation Error")


class MemoryErrorDialog(ErrorDialog):
    def __init__(self, parent, message):
        """
        Error dialog using custom MemoryErrorDialog class
        :param parent: the wx parent object
        """
        ErrorDialog.__init__(self, parent, message, "Memory Error")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# models.dvh.py
"""
Class to retrieve DVH data from SQL, calculate parameters dependent on DVHs, and extract plotting data
"""
# Copyright (c) 2016-2019 Dan Cutright
# This file is part of DVH Analytics, released under a BSD license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVH-Analytics

from copy import deepcopy
from dateutil.parser import parse as date_parser
import numpy as np
from dvha.db.sql_connector import DVH_SQL
from dvha.db.sql_to_python import QuerySQL
import operator as operator_lib
from os import listdir
from os.path import join, basename, splitext
from dvha.options import Options
from dvha.paths import PROTOCOL_DIR


MAX_DOSE_VOLUME = Options().MAX_DOSE_VOLUME


# This class retrieves DVH data from the SQL database and calculates statistical DVHs (min, max, quartiles)
# It also provides some inspection tools of the retrieved data
class DVH:
    def __init__(self, uid=None, dvh_condition=None, dvh_bin_width=5):
        """
        This class will retrieve DVHs and other data in the DVH SQL table meeting the given constraints,
        it will also parse the DVH_string into python lists and retrieve the associated Rx dose
        :param uid: a list of allowed study_instance_uids in data set
        :param dvh_condition: a string in SQL syntax applied to a DVH Table query
        :param dvh_bin_width: retrieve every nth value from dvh_string in SQL
        :type dvh_bin_width: int
        """

        self.dvh_bin_width = dvh_bin_width

        if uid:
            constraints_str = "study_instance_uid in ('%s')" % "', '".join(uid)
            if dvh_condition:
                constraints_str = "(%s) and %s" % (dvh_condition, constraints_str)
        else:
            constraints_str = ''

        # Get DVH data from SQL and set as attributes
        dvh_data = QuerySQL('DVHs', constraints_str)
        if dvh_data.mrn:
            ignored_keys = {'cnx', 'cursor', 'table_name', 'constraints_str', 'condition_str'}
            self.keys = []
            for key, value in dvh_data.__dict__.items():
                if not key.startswith("__") and key not in ignored_keys:
                    if key == 'dvh_string':
                        dvh_split = [dvh.split(',')[::self.dvh_bin_width] for i, dvh in enumerate(value)]
                    setattr(self, key, value)
                    if '_string' not in key:
                        self.keys.append(key)

            # Move mrn to beginning of self.keys
            if 'mrn' in self.keys:
                self.keys.pop(self.keys.index('mrn'))
                self.keys.insert(0, 'mrn')

            # uid passed into DVH init is a unique list for querying and it is not in order of query results
            self.uid = self.study_instance_uid

            # Add these properties to dvh_data since they aren't in the DVHs SQL table
            self.count = len(self.mrn)
            self.study_count = len(set(self.uid))
            self.rx_dose = self.get_plan_values('rx_dose')
            self.sim_study_date = self.get_plan_values('sim_study_date')
            self.keys.append('rx_dose')
            self.endpoints = {'data': None,
                              'defs': None}
            self.eud = None
            self.ntcp_or_tcp = None

            self.bin_count = max([len(dvh) for dvh in dvh_split])

            self.dvh = np.zeros([self.bin_count, self.count])

            # Get needed values not in DVHs table
            for i in range(self.count):
                # Process dvh_string to numpy array, and pad with zeros at the end
                # so that all dvhs are the same length
                current_dvh = np.array(dvh_split[i], dtype='|S4').astype(np.float)
                current_dvh_max = np.max(current_dvh)
                if current_dvh_max > 0:
                    current_dvh = np.divide(current_dvh, current_dvh_max)
                zero_fill = np.zeros(self.bin_count - len(current_dvh))
                self.dvh[:, i] = np.concatenate((current_dvh, zero_fill))

            self.dth = []
            for i in range(self.count):
                # Process dth_string to numpy array
                try:
                    self.dth.append(np.array(self.dth_string[i].split(','), dtype='|S4').astype(np.float))
                except Exception:
                    self.dth.append(np.array([0]))

            # Store these now so they can be saved in DVH object without needing to query later
            with DVH_SQL() as cnx:
                self.physician_count = len(cnx.get_unique_values('Plans', 'physician',
                                                                 "study_instance_uid in ('%s')" % "','".join(self.uid)))
            self.total_fxs = self.get_plan_values('fxs')
            self.fx_dose = self.get_rx_values('fx_dose')
        else:
            self.count = 0

        self.protocols = Protocols()

    def get_plan_values(self, plan_column):
        """
        Get values from the Plans table and store in order matching mrn / study_instance_uid
        :param plan_column: name of the SQL column to be queried
        :type plan_column: str
        :return: values from the Plans table for the DVHs stored in this class
        :rtype: list
        """
        with DVH_SQL() as cnx:
            condition = "study_instance_uid in ('%s')" % "','".join(self.study_instance_uid)
            data = cnx.query('Plans', 'study_instance_uid, %s' % plan_column, condition)
            force_date = cnx.is_sqlite_column_datetime('Plans', plan_column)  # returns False for pgsql

        uids = [row[0] for row in data]
        values = [row[1] for row in data]
        if force_date:  # sqlite does not have date or time like variables
            for i, value in enumerate(values):
                try:
                    if type(value) is int:
                        values[i] = str(date_parser(str(value)))
                    else:
                        values[i] = str(date_parser(value))
                except Exception:
                    values[i] = 'None'
        return [values[uids.index(uid)] for uid in self.study_instance_uid]

    def get_rx_values(self, rx_column):
        """
        Get values from the Rxs table and store in order matching mrn / study_instance_uid
        :param rx_column: name of the SQL column to be queried
        :type rx_column: str
        :return: values from the Rxs table for the DVHs stored in this class
        :rtype: list
        """
        with DVH_SQL() as cnx:
            condition = "study_instance_uid in ('%s')" % "','".join(self.study_instance_uid)
            data = cnx.query('Rxs', 'study_instance_uid, %s' % rx_column, condition)

        uids = [row[0] for row in data]
        values = [row[1] for row in data]
        final_values = {}
        for i, uid in enumerate(uids):
            if uid in list(values):
                final_values[uid] = "%s,%s" % (final_values[uid], values[i])
            else:
                final_values[uid] = str(values[i])
        return [final_values[uid] for uid in self.study_instance_uid]

    @property
    def x_data(self):
        """
        Get x-values of the DVHs.  All DVHs stored in SQL database are 1cGy binned csv strings
        :return: x data for plotting
        :rtype: list
        """
        return [np.multiply(np.array(range(self.bin_count)), self.dvh_bin_width).tolist()] * self.count

    @property
    def y_data(self):
        """
        Get y-values of the DVHs
        :return: all DVHs in order (i.e., same as mrn, study_instance_uid)
        :rtype: list
        """
        return [self.dvh[:, i].tolist() for i in range(self.count)]

    def get_cds_data(self, keys=None):
        """
        Get data from this class in a format compatible with bokeh's ColumnDataSource.data
        :param keys: optionally specify which properties to in include
        :return: data from this class
        :rtype: dict
        """
        if not keys:
            keys = self.keys

        return deepcopy({key: getattr(self, key) for key in keys})

    def get_percentile_dvh(self, percentile):
        """
        :param percentile: the percentile to calculate for each dose-bin
        :return: a single DVH such that each bin is the given percentile of each bin over the whole sample
        :rtype: numpy 1D array
        """
        return np.percentile(self.dvh, percentile, 1)

    def get_dose_to_volume(self, volume, volume_scale='absolute', dose_scale='absolute'):
        """
        :param volume: the specified volume in cm^3
        :param volume_scale: either 'relative' or 'absolute'
        :param dose_scale: either 'relative' or 'absolute'
        :return: the dose in Gy to the specified volume
        :rtype: list
        """
        doses = np.zeros(self.count)
        for x in range(self.count):
            dvh = np.zeros(len(self.dvh))
            for y in range(len(self.dvh)):
                dvh[y] = self.dvh[y][x]
            if volume_scale == 'relative':
                doses[x] = dose_to_volume(dvh, volume, dvh_bin_width=self.dvh_bin_width)
            else:
                if self.volume[x]:
                    doses[x] = dose_to_volume(dvh, volume/self.volume[x], dvh_bin_width=self.dvh_bin_width)
                else:
                    doses[x] = 0
        if dose_scale == 'relative':
            if self.rx_dose[0]:
                doses = np.divide(doses * 100, self.rx_dose[0:self.count])
            else:
                self.rx_dose[0] = 1  # if review dvh isn't defined, the following line would crash
                doses = np.divide(doses * 100, self.rx_dose[0:self.count])
                self.rx_dose[0] = 0
                doses[0] = 0

        return doses.tolist()

    def get_volume_of_dose(self, dose, dose_scale='absolute', volume_scale='absolute'):
        """
        :param dose: input dose use to calculate a volume of dose for entire sample
        :param dose_scale: either 'absolute' or 'relative'
        :param volume_scale: either 'absolute' or 'relative'
        :return: a list of V_dose
        :rtype: list
        """
        volumes = np.zeros(self.count)
        for x in range(self.count):

            dvh = np.zeros(len(self.dvh))
            for y in range(len(self.dvh)):
                dvh[y] = self.dvh[y][x]
            if dose_scale == 'relative':
                if isinstance(self.rx_dose[x], str):
                    volumes[x] = 0
                else:
                    volumes[x] = volume_of_dose(dvh, dose * self.rx_dose[x], dvh_bin_width=self.dvh_bin_width)
            else:
                volumes[x] = volume_of_dose(dvh, dose, dvh_bin_width=self.dvh_bin_width)

        if volume_scale == 'absolute':
            volumes = np.multiply(volumes, self.volume[0:self.count])
        else:
            volumes = np.multiply(volumes, 100.)

        return volumes.tolist()

    def coverage(self, rx_dose_fraction):
        """
        :param rx_dose_fraction: relative rx dose to calculate fractional coverage
        :return: fractional coverage
        :rtype: list
        """

        answer = np.zeros(self.count)
        for x in range(self.count):
            answer[x] = self.get_volume_of_dose(float(self.rx_dose[x] * rx_dose_fraction))

        return answer

    def get_resampled_x_axis(self):
        """
        :return: the x axis of a resampled dvh
        """
        x_axis, dvhs = self.resample_dvh()
        return x_axis

    def get_stat_dvh(self, stat_type='mean', dose_scale='absolute', volume_scale='relative'):
        """
        :param stat_type: either min, mean, median, max, or std
        :param dose_scale: either 'absolute' or 'relative'
        :param volume_scale: either 'absolute' or 'relative'
        :return: a single dvh where each bin is the stat_type of each bin for the entire sample
        :rtype: numpy 1D array
        """
        if dose_scale == 'relative':
            x_axis, dvhs = self.resample_dvh()
        else:
            dvhs = self.dvh

        if volume_scale == 'absolute':
            dvhs = self.dvhs_to_abs_vol(dvhs)

        stat_function = {'min': np.min,
                         'mean': np.mean,
                         'median': np.median,
                         'max': np.max,
                         'std': np.std}
        dvh = stat_function[stat_type](dvhs, 1)

        return dvh

    def get_standard_stat_dvh(self, dose_scale='absolute', volume_scale='relative'):
        """
        :param dose_scale: either 'absolute' or 'relative'
        :param volume_scale: either 'absolute' or 'relative'
        :return: a standard set of statistical dvhs (min, q1, mean, median, q1, and max)
        :rtype: dict
        """
        if dose_scale == 'relative':
            x_axis, dvhs = self.resample_dvh()
        else:
            dvhs = self.dvh

        if volume_scale == 'absolute':
            dvhs = self.dvhs_to_abs_vol(dvhs)

        standard_stat_dvh = {'min': np.min(dvhs, 1),
                             'q1': np.percentile(dvhs, 25, 1),
                             'mean': np.mean(dvhs, 1),
                             'median': np.median(dvhs, 1),
                             'q3': np.percentile(dvhs, 75, 1),
                             'max': np.max(dvhs, 1)}

        return standard_stat_dvh

    def dvhs_to_abs_vol(self, dvhs):
        """
        :param dvhs: relative DVHs (dvh[bin, roi_index])
        :return: absolute DVHs
        :rtype: numpy 2D array
        """
        return np.multiply(dvhs, self.volume)

    def resample_dvh(self, resampled_bin_count=5000):
        """
        :return: x-axis, y-axis of resampled DVHs
        """

        min_rx_dose = np.min(self.rx_dose) * 100.
        new_bin_count = int(np.divide(float(self.bin_count), min_rx_dose) * resampled_bin_count)

        x1 = np.linspace(0, self.bin_count, self.bin_count)
        y2 = np.zeros([new_bin_count, self.count])
        for i in range(self.count):
            x2 = np.multiply(np.linspace(0, new_bin_count, new_bin_count),
                             self.rx_dose[i] * 100. / resampled_bin_count)
            y2[:, i] = np.interp(x2, x1, self.dvh[:, i])
        x2 = np.divide(np.linspace(0, new_bin_count, new_bin_count), resampled_bin_count)
        return x2, y2

    def get_summary(self):
        summary = ["Study count: %s" % self.study_count,
                   "DVH count: %s" % self.count,
                   "Institutional ROI count: %s" % len(set(self.institutional_roi)),
                   "Physician ROI count: %s" % len(set(self.physician_roi)),
                   "ROI type count: %s" % len(set(self.roi_type)),
                   "Physician count: %s" % self.physician_count,
                   "\nMin, Mean, Max",
                   "Rx dose (Gy): %0.2f, %0.2f, %0.2f" % (min(self.rx_dose),
                                                          sum(self.rx_dose) / self.count,
                                                          max(self.rx_dose)),
                   "Volume (cc): %0.2f, %0.2f, %0.2f" % (min(self.volume),
                                                         sum(self.volume) / self.count,
                                                         max(self.volume)),
                   "Min dose (Gy): %0.2f, %0.2f, %0.2f" % (min(self.min_dose),
                                                           sum(self.min_dose) / self.count,
                                                           max(self.min_dose)),
                   "Mean dose (Gy): %0.2f, %0.2f, %0.2f" % (min(self.mean_dose),
                                                            sum(self.mean_dose) / self.count,
                                                            max(self.mean_dose)),
                   "Max dose (Gy): %0.2f, %0.2f, %0.2f" % (min(self.max_dose),
                                                           sum(self.max_dose) / self.count,
                                                           max(self.max_dose))]
        return '\n'.join(summary)

    @property
    def has_data(self):
        return bool(len(self.mrn))

    def evaluate_constraint(self, index, constraint):
        """

        :param index: The index of the dvh to be evaluated (self.dvh[:, index])
        :type index: int
        :param constraint: dvh endpoint constraint
        :type constraint: Constraint
        :return: True if the dvh meets the constraint
        :rtype: bool
        """
        return constraint(self.dvh[:, index], self.mean_dose[index], self.volume[index], self.dvh_bin_width)

    def evaluate_dvh(self, index, protocol_roi, protocol_name, fractionation):
        """

        :param index: The index of the dvh to be evaluated (self.dvh[:, index])
        :type index: int
        :param protocol_roi: the name of the roi as defined in the protocol
        :type protocol_roi: str
        :param protocol_name: name of the protocol (e.g., TG101, PEMBRO)
        :type protocol_name: str
        :param fractionation: number of fractions associated with constraints, str(int) or 'Std'
        :type fractionation: str
        :return: evaluation of all constraints for the protocol_roi with provided dvh
        :rtype: dict
        """
        constraints = self.protocols.get_roi_constraints(protocol_roi, protocol_name, fractionation,
                                                         roi_type=self.roi_type[index])
        return {str(c): self.evaluate_constraint(index, c) for c in constraints}


# Returns the isodose level outlining the given volume
def dose_to_volume(dvh, rel_volume, dvh_bin_width=1):
    """
    :param dvh: a single dvh
    :param rel_volume: fractional volume
    :param dvh_bin_width: dose bin width of dvh
    :type dvh_bin_width: int
    :return: minimum dose in Gy of specified volume
    """

    # Return the maximum dose instead of extrapolating
    if rel_volume < dvh[-1]:
        return len(dvh) * dvh_bin_width * 0.01

    dose_high = np.argmax(dvh < rel_volume)
    y = rel_volume
    x_range = [dose_high - 1, dose_high]
    y_range = [dvh[dose_high - 1], dvh[dose_high]]
    dose = np.interp(y, y_range, x_range) * dvh_bin_width * 0.01

    return dose


def volume_of_dose(dvh, dose, dvh_bin_width=1):
    """
    :param dvh: a single dvh
    :param dose: dose in cGy
    :param dvh_bin_width: dose bin width of dvh
    :type dvh_bin_width: int
    :return: volume in cm^3 of roi receiving at least the specified dose
    """

    x = [int(np.floor(dose / dvh_bin_width * 100)), int(np.ceil(dose / dvh_bin_width * 100))]
    if len(dvh) < x[1]:
        return dvh[-1]
    y = [dvh[x[0]], dvh[x[1]]]
    roi_volume = np.interp(float(dose) / dvh_bin_width, x, y)

    return roi_volume


def calc_eud(dvh, a, dvh_bin_width=1):
    """
    EUD = sum[ v(i) * D(i)^a ] ^ [1/a]
    :param dvh: a single DVH as a list of numpy 1D array with 1cGy bins
    :param a: standard a-value for EUD calculations, organ and dose fractionation specific
    :param dvh_bin_width: dose bin width of dvh
    :type dvh_bin_width: int
    :return: equivalent uniform dose
    :rtype: float
    """
    v = -np.gradient(dvh)

    dose_bins = np.linspace(0, np.size(dvh), np.size(dvh))
    dose_bins = np.round(dose_bins, 3) * dvh_bin_width
    bin_centers = dose_bins - (dvh_bin_width / 2.)
    eud = np.power(np.sum(np.multiply(v, np.power(bin_centers, a))), 1. / float(a))
    eud = np.round(eud, 2) * 0.01

    return eud


def calc_tcp(gamma_50, td_tcd, eud):
    return 1. / (1. + (td_tcd / eud) ** (4. * gamma_50))


class Protocols:
    """Parse all protocols in user's PROTOCOL_DIR"""
    def __init__(self, max_dose_volume=MAX_DOSE_VOLUME):
        self.max_dose_volume = max_dose_volume

        # Collect .scp files, parse, store in self.data
        self.data = {}
        for f in self.file_paths:
            file_name = splitext(str(basename(f)))[0]
            protocol_name, fxs = tuple(file_name.split('_'))
            if protocol_name not in list(self.data):
                self.data[protocol_name] = {}
            fxs_key = fxs.lower().replace('fx', '').strip()
            self.data[protocol_name][fxs_key] = self.parse_protocol_file(f)

    @property
    def file_paths(self):
        files = listdir(PROTOCOL_DIR)
        return [join(PROTOCOL_DIR, f) for f in files if self.is_protocol_file(f) and f.count('_') == 1]

    @staticmethod
    def is_protocol_file(file_path):
        return splitext(file_path)[1].lower() == '.scp'

    @staticmethod
    def parse_protocol_file(file_path):
        constraints = {}
        current_key = None
        with open(file_path, 'r') as document:
            for line in document:
                if not(line.startswith('#') or line.strip() == ''):  # Skip line if empty or starts with #
                    if line[0] not in {'\t', ' '}:  # Constraint
                        current_key = line.strip()
                        constraints[current_key] = {}
                    else:  # OAR Name
                        line_data = line.split()
                        operator = '<' if len(line_data) < 3 else line_data[-2]
                        # constraints[current_key][line_data[0]] = {'threshold': line_data[-1],
                        #                                           'operator': operator}
                        constraints[current_key][line_data[0]] = Constraint(line_data[0], operator, line_data[-1])
        return constraints

    @property
    def protocol_names(self):
        return sorted(list(self.data))

    def get_fractionations(self, protocol_name):
        return [fx.replace('Fx', '') for fx in sorted(list(self.data[protocol_name]))]

    def get_rois(self, protocol_name, fractionation):
        return sorted(list(self.data[protocol_name][fractionation]))

    def get_constraints(self, protocol_name, fractionation, roi_name):
        return self.data[protocol_name][fractionation][roi_name]

    def get_column_data(self, protocol_name, fractionation, roi=None):

        columns = ['Structure', 'ROI Type', 'Constraint']
        data = {key: [] for key in columns}
        rois = [roi] if roi is not None else self.get_rois(protocol_name, fractionation)
        for roi in rois:
            constraints = self.get_roi_constraints(roi, protocol_name, fractionation)
            for i, constraint in enumerate(constraints):
                if i:
                    data['Structure'].append('')
                    data['ROI Type'].append('')
                else:
                    data['Structure'].append(roi)
                    data['ROI Type'].append(constraint.roi_type)
                data['Constraint'].append(str(constraint))
            if len(constraints) and roi != rois[-1]:
                for col in columns:
                    data[col].append('')

        return data, columns

    def get_roi_constraints(self, protocol_roi, protocol_name, fractionation):
        return self.get_constraints(protocol_name, fractionation, protocol_roi).values()

    @staticmethod
    def get_roi_type(roi_name):
        roi_name = roi_name.lower()
        if any([t in roi_name for t in {'gtv', 'ctv', 'itv', 'ptv'}]):
            return 'TARGET'
        return 'OAR'

    def delete_protocol(self, protocol_name):
        if protocol_name in list(self.data):
            self.data.pop(protocol_name, None)

    def delete_fractionation(self, protocol_name, fractionation):
        if protocol_name in list(self.data) and fractionation in list(self.data[protocol_name]):
            self.data[protocol_name].pop(fractionation, None)

    def delete_constraint(self, protocol_name, fractionation, roi, constraint_label):
        try:
            self.data[protocol_name][fractionation][roi].pop(constraint_label, None)
        except KeyError:
            pass

    def add_constraint(self, protocol_name, fractionation, roi_name, constraint):
        protocol = self.data[protocol_name][fractionation]
        if roi_name not in list(protocol):
            protocol[roi_name] = {}
        protocol[roi_name][str(constraint)] = constraint

    def save_old(self, protocol, fractionation):
        file_path = join(PROTOCOL_DIR, "%s_%s.scp_test" % (protocol, fractionation))
        with open(file_path, 'w') as doc:
            lines = []
            for roi, constraints in self.data[protocol][fractionation].items():
                lines.append(roi)
                for name, threshold in constraints.items():
                    lines.append("\t%s\t<\t%s" % (name, threshold))
            doc.write('\n'.join(lines))


class Constraint:
    def __init__(self, label, operator, threshold, roi_type='OAR', max_dose_volume=MAX_DOSE_VOLUME):
        self.label = label
        self.operator = operator
        self.threshold = threshold
        self.roi_type = roi_type
        self.max_dose_volume = max_dose_volume

    def __str__(self):
        return "%s %s %s" % (self.label_with_units, self.operator, self.threshold_with_units)

    def __repr__(self):
        return self.__str__()

    def __call__(self, dvh, volume, bin_width=1, mean_dose=None):
        """
        Determine if a DVH meets this constraint
        :param dvh: a relative DVH
        :type dvh: np.array
        :param volume: ROI volume
        :type volume: float
        :param bin_width: dose distance between elements of dvh, default value is 1
        :type bin_width: int
        :param mean_dose: optionally provide ROI mean dose for efficiency
        :type mean_dose: float
        :return: the pass/fail status of the dvh with this constraint
        :rtype: bool
        """
        if mean_dose is None:  # if mean dose not provided, calculate it
            diff = abs(np.diff(np.append(dvh, 0)))  # per dicompyler-core
            mean_dose = (diff.bincenters * diff.counts).sum() / diff.counts.sum()

        calc_type = self.calc_type  # calculate only once
        if calc_type == 'Volume':
            endpoint = dose_to_volume(dvh, self.input_value, dvh_bin_width=bin_width)
        elif calc_type == 'Dose':
            endpoint = volume_of_dose(dvh, self.input_value, dvh_bin_width=bin_width)
        elif calc_type == 'Mean':
            endpoint = mean_dose
        elif calc_type == 'MVS':
            endpoint = volume - volume_of_dose(dvh, self.input_value, dvh_bin_width=bin_width)
        else:
            return

        func = operator_lib.lt if self.operator == '<' else operator_lib.gt
        return func(endpoint, self.threshold_value)

    @property
    def threshold_value(self):
        if '%' in self.threshold:
            return float(self.threshold.replace('%', '')) / 100.
        return float(self.threshold)

    @property
    def output_type(self):
        if self.label == 'Mean':
            return 'D'
        return self.label.split('_')[0]

    @property
    def output_units(self):
        return ['Gy', 'cc'][self.output_type in {'V', 'MVS'}]

    @property
    def input(self):
        if self.label == 'Mean':
            return None
        return self.label.split('_')[1]

    @property
    def input_type(self):
        return ['Volume', 'Dose'][self.output_type in {'V', 'MVS'}]

    @property
    def calc_type(self):
        if 'MVS' in self.label:
            return 'MVS'
        if 'Mean' in self.label:
            return 'Mean'
        return self.input_type

    @property
    def input_value(self):
        if self.input is None:
            return None
        if 'max' in self.input:
            return self.max_dose_volume
        return float(self.input.replace('%', '').replace('_', ''))

    @property
    def input_scale(self):
        if self.input is None:
            return None
        return ['absolute', 'relative']['%' in self.input]

    @property
    def output_scale(self):
        return ['absolute', 'relative']['%' in self.threshold]

    @property
    def input_units(self):
        if self.input is None:
            return None
        scale = ['absolute', 'relative']['%' in self.input]
        abs_units = ['cc', 'Gy'][self.input_type == 'Dose']
        return ['%', abs_units][scale == 'absolute']

    @property
    def label_with_units(self):
        input_units = self.input_units
        if '%' not in self.label and not self.label.lower().endswith('max') and input_units:
            return self.label + input_units
        return self.label

    @property
    def threshold_with_units(self):
        if '%' not in self.threshold:
            return self.threshold + self.output_units
        return self.threshold
